/* GDFIO
      data structures and File IO
*/

#ifndef LENDEFN

#define LENDEFN      9  /* length of definition names (8=DOS filename limit) + null char */
#define LENDESC     41  /* 101 needed; length of descriptions in libraries */
#define QTYALLOWED 100  /* number of occurrences allowed */
#define LENRT        5  /* length of record_types (4) + null char */

   /* length of acronym item + null char */
#define LENACRO      9  /* labels (8=specification) */
#define LENFORMAT    9  /* format limit */
#define LENNAME     41  /* full_name limit */
#define LENUNIT     21  /* units limit */
#define LENCOMMENT 101  /* comment limit */

   /* limits on internal data set sizes */
#define  LIBLIMIT   10  /* libraries opened */
#define DEFNLIMIT  500  /* total definitions of all open libraries */
#define   RTLIMIT 1000  /* total record types */
#define ACROLIMIT 2000  /* total acronyms */
#define     LIMIT 2000  /* max of RTLIMIT and ACROLIMIT */

#define MAXDEFNLINE 8000  /* maximum size of a DEFN record line */
int defnLineNumber =  0;  /* definition_record rt line index */
int focused      = true;  /* true if only subset R/A shown for D/R lists */

    int DEFNPAGE =  9;  /* quantity of definitions to be displayed */
    int RTPAGE   = 13;  /* quantity of record_types to be displayed */
    int ACROPAGE = 17;  /* quantity of acronyms to be displayed */

struct DEFN {
   /* stored */
      int  source;               /* source library index */
      char defn[LENDEFN];        /* abbreviation for defn = filename */
      char description[LENDESC]; /* for library identification */
      int  RTused;               /* quantity of RTid used */
      int  RTid[QTYALLOWED];     /* record_types used, as index numbers */

   /* used in program */
      int  RTnull;               /* index of zero-length RT, = -1 if none */
      struct DEFN *next;         /* chain down the definitions */
      struct DEFN *prev;         /* chain up the definitions */
} *Defn, *DefnRoot, *DefnTail, *DefnActive;

struct RECTYPE {
   /* stored */
      int  source;               /* source library */
      char rt[LENRT];            /* record_type name */
      char description[LENDESC]; /* for library identification */
      int  Aused;                /* quantity of Aid used */
      int  Aid[QTYALLOWED];      /* acronyms used as index numbers */

   /* used in program */
      struct RECTYPE *reference; /* holds destination record type for format convertions */
      int  order;                /* when saving Libraries, maintains RTid
                                    when displaying files, holds record size */
} *RT;

struct ACRONYM {
   /* stored */
      int  source;               /* source library */
      char acronym[LENACRO];     /* acronym of entity */
/*      long start;                /* start of a set of readings */
      char format[LENFORMAT];    /* FORTRAN format */
      char name[LENNAME];        /* full name of acronym entity */
      char unit[LENUNIT];        /* entity units */
    double null;                 /* value used for null readings */
      char comment[LENCOMMENT];  /* for DEFN user information */

   /* used in program */
      int  fwidth;               /* holds format width */
      int  fstart;               /* holds format start position in data record */
      char Cformat[LENFORMAT];   /* holds C format equivalent */
      struct ACRONYM *reference; /* holds source acronym in format convertions */
      int  order;                /* when saving Libraries, maintains Aid
                                    when displaying files, holds display width */
} *Acro,*AcroS;

struct { /* holds information on the lists used by
            definitions, record_types, and acronyms, ie the DRA lists
            list has <page> entries from one of the DRA (the selection) */
      int used;   /* used rows of selected DRA */
      int displayed;/* displayed rows of selected DRA */
      int limit;  /* maximum allowed rows in selected DRA */
      int page;   /* max quantity of rows from selected DRA to be displayed */
      int start;  /* index to selected DRA of first displayed entry */
      int row;    /* currently active row as selected DRA index */
      int column; /* currently active column */
} Lt[3];

char   sourceName[LIBLIMIT][PATHSIZE];   /* holds library names */

   /* the database sets */
struct RECTYPE   *RT_ptr[RTLIMIT];
struct ACRONYM   *Acronym_ptr[ACROLIMIT];
   /* structures to keep copies of DRA entries */
struct DEFN      DefnKept;
struct RECTYPE   RTKept;
struct ACRONYM   AcroKept;

   /* quantity in sets that have been used */
int maxLibraries = 0;
int maxRecType   = 0;
int maxAcronym   = 0;

int activeLibrary;
int active[3]   = {0,0,0};  /* the active DRA */
int selected[3] = {0,0,0};  /* true after a DRA Copy */

FILE *stream;        /* for general IO use */
FILE *sourceStream;  /* for source file IO use */
FILE *destStream;    /* for destination file IO use */


/* prototypes */


void NullDRA(void);
     /* initiallize the DRA data sets */
void SetWorkPath(void);
     /* set path to user requested directory */
int  OpenFile(char *path, char *access, int action, char useFile);
     /* opens a file for access via the stream selected by <useFile> */
void SaveLib(int source, char *comment);
     /* saves the library associated with <source> */
void LocateActiveDefn(int row);
     /* find DEFN for row */
void AddToTailDefn(struct DEFN *Defn);
     /* add to tail of DEFN list */
struct DEFN *AllocDefn(int row);
     /* make a new copy of DEFN and splice into the list at <row> */
void ReadAcro(char *acro, int len);
     /* read next value into <acro> field with max length <len> */
void GetLib(char form);
     /* loads a library */
void MakeDefn(int defnNumb);
     /* makes a DEFN file for use in data transfers */
void GetDefn(char *name);
     /* loads a DEFN file of the type used in data transfers
        if <name>="", ask for name */
void DeleteDefn(void);
     /* delete a DEFN file from disk */

void DiscardAll(void);
     /* clear all sets ready to load/write new definitions */
int  NewRow(int list, int row);
     /* determines the row (= index into one of the DRA sets) number */
int  DefnRelist(int row, int column);
     /* re-display the definition list */
void DefnListInit(int page);
     /* initiallize the definition list */
int  DefnNew(int source, int row, int column);
     /* allocate and fill a new definition */
int  RTRelist(int row, int column);
     /* re-display the record_type list */
void RTListInit(int page);
     /* initiallize the record_type list */
int  RTnew(int row, int column);
     /* allocate and fill a new record_type */
int  AcroRelist(int row, int column);
     /* re-display the acronym list */
void AcroListInit(int page);
     /* initiallize the acronym list */
int  AcroNew(int source,int row, int column);
     /* allocate and fill a new acronym */
void CheckDecimals(char *format, char *addf);
     /* fixes decimal-type C-formats */
void FormatWidths(void);
     /* determine the width of each format in the active definition */
int  MoveDRAlistRow(int list, int row);
     /* move the active row of DRA list */
int  MoveDRAlistColumn(int list, int column);
     /* move the active column of DRA list */
int  UpdateDRA(int list, int row, int column);
     /* update the information in the DRA list */
void CopyDRA(int list);
     /* keeps a copy of the active DRA entry */
void PasteDRA(int list);
     /* overwrites the active DRA entry with the kept entry */
void DRAins(int list, int row);
     /* insert a record into a DRA list */
struct DEFN *DisposeDefn(struct DEFN *DefnA);
     /* disposes of DefnA, returns next Defn */
void DRAdel(int list, int row);
     /* delete a record from a DRA list */

int  alertError(int error, char *message, char *second);
     /* displays an alert then returns the given error code */

int  defnBeheader(char *Aline, int semi, int *RTnumb, int warnRT);
     /* extracts and interprets a DEFN header */
int  whichAttrib(char *sect, int *In);
     /* determines which acronym attribute is present in string <sect>
        starting at the <In>'th character
        <In> returns as the index to the start of the information in <sect> */
int  testEqual(char *info, char *source, int I, int cut);
     /* returns information content of type <source> in <info> */
int  acroCracker(char *Asection, int RTnumb);
     /* interprets acronym section and saves in data sets */


/* implimentation */


void NullDRA(void)
     /* initiallize the DRA data sets */
{     Defn = NULL;
      DefnRoot = NULL;
      DefnTail = NULL;
      DefnActive = NULL;

      maxLibraries = 0;
      maxRecType   = 0;
      maxAcronym   = 0;

      sourceName[0][0]= '\0';
      RT_ptr[0]    = NULL;
      Acronym_ptr[0] = NULL;

      activeLibrary  = 0;
      saved = true;
}  /* end NullDRA */

void SetWorkPath(void)
     /* set path to user requested directory */
{   char path[PATHSIZE];
    int  len;
    if (strcmp(workpath,""))       /* notify user of current path */
      sprintf(path,"Working directory is currently %s",workpath);
    else
      strcpy(path,"");

    strcpy(path,alert('S',path,"Please enter your working directory : "));
    strRemoveDspace(path);         /* in case multiple spaces entered */
    if (strcmp(path," ") == 0)     /* space means make = current dir */
      strcpy(workpath,"");
    else if (strcmp(path,"")) {    /* 'Return' key = no change */
      strcpy(workpath,path);
      len = strlen(workpath);
      if ((len > 0) && (workpath[len-1] != DIRSPLITTER)) {
        workpath[len]   = DIRSPLITTER;
        workpath[len+1] = '\0';
      }
    }
}  /* end SetWorkPath */

int OpenFile(char *path, char *RWaccess, int action, char useFile)
     /* opens a file for access via the stream selected by <useFile>
       <path>     is the file path and name
       <RWaccess> is the access type eg 'r' or 'w'
       <action> = 1 when need user warning
                = 2 when overwrite if exists
       <useFile>= IO stream to open */
{   char fullpath[PATHSIZE];
    int overwrite = 1;  /* default for new files */

    if (strpos(path,DIRSPLITTER) == -1) {  /* add workpath */
      strcpy(fullpath,workpath);
      strcat(fullpath,path);
    }
    else
      strcpy(fullpath,path);
       /* check if overwriting an existing file */
    if (action < 2) {
      if ((strpos(RWaccess,'w') > -1) && (access(fullpath,0) == 0)) overwrite = 0;
      if (!overwrite)
        overwrite = alertYesNo("File already exists do you want to overwrite",fullpath);
    }
       /* open file, overwrite = OK to open or create */
    if (overwrite) {
      switch (useFile) {
        case 'S': if (sourceStream != NULL)                 /* release any handles */
                    fclose(sourceStream); sourceStream = NULL;
                  sourceStream = fopen(fullpath, RWaccess); /* open file */
                  if (sourceStream != NULL) return 1;       /* success */
                  break;
        case 'D': if (destStream != NULL)
                    fclose(destStream); destStream = NULL;
                  destStream = fopen(fullpath, RWaccess);
                  if (destStream != NULL) return 1;
                  break;
        default : if (stream != NULL)
                    fclose(stream); stream = NULL;
                  stream = fopen(fullpath, RWaccess);
                  if (stream != NULL) return 1;
      }
         /* file not successfully opened */
      if (action) alert('A',"Cannot open file ",fullpath);
    }
    return 0;
}  /* end OpenFile */


void SaveLib(int source, char *comment)
     /* saves the library associated with <source>
        <comment> is the message to the user if needed
                first char of <comment> = ' ' for SAVE AS */
{ int L,R,A;                 /* local indexes for rt and acronyms */
  int sourceCount[LIBLIMIT]; /* holds count of quantities in all libraries */
  static int IsIn[LIMIT];    /* true if record_type/acronym used by a Defn */
  int Rcount;                /* counts the record_types in internal library */
  int Acount;                /* counts the acronyms in internal library */
  char libname[33];          /* the library name */
  int SaveAll = false;       /* true if all data to be condensed into one library */
     /* ask user for a library name if one needed */
  if (source < 0) { source = 0; SaveAll = true; }
  if ((SaveAll) || (sourceName[source][0] == '\0') || (comment[0] == ' ')) {
    if (strlen(comment) == 0)
      strcpy(comment,"No LIBRARY name given");
    strcpy(sourceName[source],alert('S',comment,"Please enter a name : "));
    if ((!SaveAll) && (maxLibraries > 1)
                   && (strlen(sourceName[source]) != 0))
      SaveAll = alertYesNo("Should all definitions be saved in this Library",
                           "irrespective of original source library ?");
  }
     /* if name given, save the library */
  if (strlen(sourceName[source]) != 0) {
    if (maxLibraries <= source) maxLibraries = source+1; /* started empty */
    sprintf(libname,"%s%s",sourceName[source],LIBEXTN);
    if (OpenFile(libname,"w\0",1,' ')) {
         /* flag all unused RT */
      for (R = 0; R < maxRecType; R++) IsIn[R] = 0;
      for (Defn = DefnRoot; Defn != NULL; Defn = Defn->next) {
        for (R = 0; R < Defn->RTused; R++) IsIn[Defn->RTid[R]] = 1;
      }
      for (R = 0; R < maxRecType; R++) {
        if (IsIn[R] == 0) {
          RT = RT_ptr[R];
          RT->source = -1;
        }
      }
         /* flag all unused Acro */
      for (A = 0; A < maxAcronym; A++) IsIn[A] = 0;
      for (R = 0; R < maxRecType; R++) {
        RT = RT_ptr[R];
        if (RT->source > -1)
          for (A = 0; A < RT->Aused; A++) IsIn[RT->Aid[A]] = 1;
      }
      for (A = 0; A < maxAcronym; A++) {
        if (IsIn[A] == 0) {
          Acro = Acronym_ptr[A];
          Acro->source = -1;
        }
      }
         /* count active items */
      if (SaveAll) {  /* all entries into one library */
        Rcount = maxRecType;
        Acount = maxAcronym;
           /* redirect all defns to same lib */
        for (Defn = DefnRoot; Defn != NULL; Defn = Defn->next)
          Defn->source = source;
        for (R = 0; R < maxRecType; R++) { /* redirect all rt to lib */
          RT = RT_ptr[R];
          if (RT->source > -1) {
            RT->order  = R;
            RT->source = source;
          }
        }
        for (A = 0; A < maxAcronym; A++) { /* redirect all acronyms to lib */
          Acro = Acronym_ptr[A];
          if (Acro->source > -1) {
            Acro->order  = A;
            Acro->source = source;
          }
        }}
      else {  /* only those belonging to library source into library
                 order source indexes to maintain self-reference in file */
        for (L = 0; L < maxLibraries; L++) sourceCount[L] = -1;
        for (R = 0; R < maxRecType; R++) {
          RT = RT_ptr[R];
          sourceCount[RT->source]++;
          RT->order = sourceCount[RT->source];
        }
        Rcount = sourceCount[source]+1;

        for (L = 0; L < maxLibraries; L++) sourceCount[L] = -1;
        for (A = 0; A < maxAcronym; A++) {
          Acro = Acronym_ptr[A];
          sourceCount[Acro->source]++;
          Acro->order = sourceCount[Acro->source];
        }
        Acount = sourceCount[source]+1;
      }
      /* write library file */
         /* counters for checking if can re-load */
      fprintf(stream,"%i %i\n",Rcount,Acount);
         /* dump definition list belonging to source library */
      for (Defn = DefnRoot; Defn != NULL; Defn = Defn->next) {
        if (Defn->source == source) {
          fprintf(stream,"%s\n",Defn->defn);
          fprintf(stream,"%s\n",Defn->description);
          fprintf(stream,"%i ",Defn->RTused);
          for (R = 0; R < Defn->RTused; R++) {
            RT = RT_ptr[Defn->RTid[R]];
            fprintf(stream,"%i ",RT->order);
          }
          fprintf(stream,"\n");
        }
      }
      fprintf(stream,"%s\n","END DEFN");  /* trigger for start RT */
         /* dump record_type list belonging to source library */
      for (R = 0; R < maxRecType; R++) {
        RT = RT_ptr[R];
        if (RT->source == source) {
          fprintf(stream,"%s\n",RT->rt);
          fprintf(stream,"%s\n",RT->description);
          fprintf(stream,"%i ",RT->Aused);
          for (A = 0; A < RT->Aused; A++) {
            Acro = Acronym_ptr[RT->Aid[A]];
            fprintf(stream,"%i ",Acro->order);
          }
          fprintf(stream,"\n");
        }
      }
      fprintf(stream,"%s\n","END RT");  /* for readability */
         /* dump acronym list belonging to source library */
      for (A = 0; A < maxAcronym; A++) {
        Acro = Acronym_ptr[A];
        if (Acro->source == source) {
          fprintf(stream,"%s\n",Acro->acronym);
          fprintf(stream,"%s\n",Acro->format);
          fprintf(stream,"%s\n",Acro->name);
          fprintf(stream,"%s\n",Acro->unit);
          fprintf(stream,"%s\n",Acro->comment);
          fprintf(stream,"%lf\n",Acro->null);
        }
      }
      fclose(stream); stream = NULL;
      saved = true;
    }
  }
}  /* end SaveLib */


void LocateActiveDefn(int row)
     /* find DEFN for row */
{   int D;
    DefnActive = DefnRoot;
    for (D = 0; D < row; D++)
      if (DefnActive != NULL) DefnActive = DefnActive->next;
    active[DEFNLIST] = row;
}  /* end LocateActiveDefn */

void AddToTailDefn(struct DEFN *Defn)
     /* add to tail of DEFN list */
{
    if (DefnTail == NULL)
      DefnRoot = Defn;
    else
      DefnTail->next = Defn;

    DefnActive = Defn;
    Defn->next = NULL;
    Defn->prev = DefnTail;
    DefnTail = Defn;
}  /* end AddToTailDefn */

struct DEFN *AllocDefn(int row)
     /* make a new copy of DEFN and splice into the list at <row> */
{   struct DEFN *Defn;
    Defn = (struct DEFN *) malloc(sizeof(struct DEFN)); /* allocate RAM */
    CheckRAM(Defn);              /* make sure enough room existed in RAM */
       /* fill with empty */
    Defn->source = maxLibraries;
    Defn->defn[0] = '\0';
    Defn->description[0] = '\0';
    Defn->RTused = 0;
    Defn->RTid[0] = 0;
    Defn->RTnull = -1;
    Lt[DEFNLIST].used++;
       /* place in Defn chain */
    if ((row < 0) || (DefnTail == NULL))  /* add to tail */
      AddToTailDefn(Defn);
    else {                       /* insert into list */
      LocateActiveDefn(row);
      if (DefnActive == NULL)
        AddToTailDefn(Defn);
      else {
        Defn->next = DefnActive;
        Defn->prev = DefnActive->prev;
        DefnActive->prev = Defn;
        if (Defn->prev == NULL)
          DefnRoot = Defn;
        else
          (Defn->prev)->next = Defn;
      }
    }
    return Defn;
}  /* end AllocDefn */


void ReadAcro(char *acro, int len)
     /* read next value into <acro> field with max length <len> */
{   char dummy[LENCOMMENT+1];
    if (fgets(dummy,100,stream) != NULL) {
      dummy[strlen(dummy)-1] = '\0';    /* remove line feed */
      dummy[len] = '\0';                /* trim any excess */
      strcpy(acro,dummy);
    }
}  /* end ReadAcro */

void GetLib(char form)
     /* loads a library
        <form> = 'g' if need to ask user for a library name
               = 'n' for new
               = 'i' for initiallization call */
{
  int  addR,addA;    /* file library contains these qty of rt and acros */
  int  R,A,N;        /* local indexes */
  char marker[16];   /* dummy string */
  char *isEof;       /* end of file indicator (corrupted file) */
  char libname[PATHSIZE]; /* library name */

  if (maxLibraries >= LIBLIMIT)
    alert('A',"Maximum number of Libraries","have already been opened");
  else {
       /* get library name from list of LIBEXTN files or user-supplied name */
    if ((form == 'g') || (form == 'i')) {
      strcpy(libname,FileList("",LIBEXTN,0));
      if (strcmp(libname,"?") == 0) {
        if (form == 'g') alert('A',"There are no Libraries to open !","");
        libname[0] = '\0';
      }}
    else
      strcpy(libname,alert('S',"Library needed","please enter a name : "));

    if (strlen(libname) > 0) {
      strcpy(sourceName[maxLibraries],libname);
      PrintCharWindow('R',1,BROWN,BLACK,sourceName[maxLibraries]);
      strcat(libname,LIBEXTN);
      if (!OpenFile(libname,"r\0",0,' '))
        alert('A',"You have just created a NEW library called ",libname);
      else if (feof(stream))
        alert('A',"You tried to load an EMPTY library called ",libname);
      else {
           /* read library file */
        fscanf(stream,"%i %i",&addR,&addA);
        fgets(marker,16,stream);  /* to skip line feed */
           /* make sure enough room exists in sets to load the library */
        if ((maxRecType+addR < RTLIMIT) && (maxAcronym+addA < ACROLIMIT)) {
             /* load the definitions into next available set */
          do {
            isEof = fgets(marker,LENDEFN+1,stream);
            if (isEof != NULL) {
              marker[strlen(marker)-1] = '\0';    /* remove line feed */
              if (strcmp(marker,"END DEFN")) {
                Defn = AllocDefn(-1);             /* add to tail */
                strcpy(Defn->defn,marker);
                isEof = fgets(Defn->description,LENDESC,stream);
                if (isEof != NULL) {
                  Defn->description[strlen(Defn->description)-1] = '\0'; /* remove line feed */
                  if (fscanf(stream,"%i",&Defn->RTused) != 1) isEof = NULL;
                  else {
                    for (R = 0; (R < Defn->RTused) && (isEof != NULL); R++) {
                      if (fscanf(stream,"%i",&N) != 1) isEof = NULL;
                      Defn->RTid[R] = N + maxRecType; /* offset to what will be the rt index */
                    };
                    isEof = fgets(marker,16,stream);    /* to skip line feed */
                  }
                }
              }
            }
          } while (strcmp(marker,"END DEFN") && (isEof != NULL));

             /* load the record_types into next available set */
          if (isEof != NULL) {
           isEof = fgets(marker,16,stream);  /* to check for END RT */
           for (R = 0; (R < addR) && (isEof != NULL); R++) {
             marker[strlen(marker)-1] = '\0';    /* remove line feed */
             if (strcmp(marker,"END RT")) {
               RT_ptr[maxRecType] = (struct RECTYPE *) malloc(sizeof(*RT));  /* allocate RAM */
               RT = RT_ptr[maxRecType];
               CheckRAM(RT);      /* make sure enough room exists in RAM */
               RT->source = maxLibraries;
               maxRecType++;
               strcpy(RT->rt,marker);
               fgets(RT->description,LENDESC,stream);
               RT->description[strlen(RT->description)-1] = '\0'; /* remove line feed */
               if (fscanf(stream,"%i",&RT->Aused) != 1) isEof = NULL;
               for (A = 0; (A < RT->Aused) && (isEof != NULL); A++) {
                 if (fscanf(stream,"%i",&N) != 1) isEof = NULL;
                 else RT->Aid[A] = N + maxAcronym;    /* offset to what will be the acronym index */
               };
               fgets(marker,16,stream);  /* to skip line feed */
             };
             isEof = fgets(marker,16,stream);    /* to get next <rt> or 'END RT' */
           };
              /* load the acronyms into next available set */
           for (A = 0; (A < addA) && (isEof != NULL); A++) {
             Acronym_ptr[maxAcronym] = (struct ACRONYM *) malloc(sizeof(*Acro)); /* allocate RAM */
             Acro = Acronym_ptr[maxAcronym];
             CheckRAM(Acro);      /* make sure enough room exists in RAM */
             Acro->source = maxLibraries;
             maxAcronym++;
             ReadAcro(Acro->acronym,LENACRO);
             ReadAcro(Acro->format,LENFORMAT);
             ReadAcro(Acro->name,LENNAME);
             ReadAcro(Acro->unit,LENUNIT);
             ReadAcro(Acro->comment,LENCOMMENT);
             fscanf(stream,"%lf",&Acro->null);
             isEof = fgets(marker,16,stream);          /* to skip line feed */
           }
          }
        }
      }
      fclose(stream); stream = NULL;
      maxLibraries++;
      RT_ptr[maxRecType] = NULL;
      Acronym_ptr[maxAcronym] = NULL;
    }
  }
}  /* end GetLib */


void MakeDefn(int defnNumb)
     /* makes a DEFN file for use in data transfers
        <defnNumb> is the defn set index
        = -1 if use DefnActive and dataDestination */
{ int I;               /* DEFN record counter for each rt*/
  int R,A;             /* local indexes */
  char comma;          /* a comma if needed */
  char defnname[33];   /* filename for DEFN */
  char rtname[LENRT];  /* output record-type name, "<>" is none ! */
  char nullstr[20];    /* to hold the str version of Acro->null */

     /* make defnname from name in set */
  if (defnNumb >= 0) LocateActiveDefn(defnNumb);
  if (DefnActive == NULL)
    alert('A',"Select a Definition !","");
  else {
       /* create filename */
    if (defnNumb >= 0)
      strcpy(defnname,DefnActive->defn);
    else
      strcpy(defnname,ValidateDOSfileName(dataDestination,'.'));
    strcat(defnname,DFNEXTN);
       /* start a new file */
    if (OpenFile(defnname,"w\0",1,' ')) {
         /* scan all the record_types associated with the definition */
      for (R = 0; R < DefnActive->RTused; R++) {
        RT = RT_ptr[DefnActive->RTid[R]];
        I = 1;
        strcpy(rtname,RT->rt);
        if (strcmp(rtname,"<>"))   /* rt on data records */
          fprintf(stream,"DEFN %i ST=RECD,RT=%s; RT: A4\n",I++,rtname);
        else                       /* no rt on data records */
          rtname[0] = '\0';
           /* scan all the acronyms associated with the record_type */
        for (A = 0; A < RT->Aused; A++) {
          Acro = Acronym_ptr[RT->Aid[A]];
             /* place each acronym on its own line for ledgibility */
          fprintf(stream,"DEFN %i ST=RECD,RT=%s; %s",I++,rtname,Acro->acronym);
             /* only write items that are required or not empty */
/*          if (Acro->start > 1) fprintf(stream," %li",Acro->start); */
          fprintf(stream,": %s: ",Acro->format);
          comma = ' ';
          if (Acro->unit[0] != '\0') {
            fprintf(stream," UNIT=%s",Acro->unit);
            comma = ',';
          }
          if (Acro->null != NOVALUE) {
            sprintf(nullstr,"%.*f",6,Acro->null);
            TrimZero(nullstr);               /* remove excess zeros */
            fprintf(stream,"%c NULL=%s",comma,nullstr);
            comma = ',';
          }
          if (Acro->name[0] != '\0') {
            fprintf(stream,"%c NAME=%s",comma,Acro->name);
            comma = ',';
          }
          if (Acro->comment[0] != '\0')
          fprintf(stream,"%c %s",comma,Acro->comment);
          fprintf(stream,"\n");
        }
        fprintf(stream,"DEFN %i ST=RECD,RT=%s; END DEFN\n",I,rtname);
      }
      fclose(stream); stream = NULL;
    }
  }
}  /* end MakeDefn */


void GetDefn(char *name)
     /* loads a DEFN file of the type used in data transfers
        if <name>="", ask for name
        DEFN will be loaded into it's own "library" */
{ int  I;                /* index */
  int  semi;             /* location of semicolon */
  int  error = 0;        /* used in error detection */
  int  RTnumb;           /* DEFN record index for record_types */
  char defnname[33];     /* filename for DEFN */
  char *Aline,*Alineptr; /* a line from the DEFN file */
  char *Asection;        /* a section of Aline to be cracked */
  Aline = (char *) malloc(MAXDEFNLINE+1);
  Asection = (char *) malloc(1001);

  if (maxLibraries >= LIBLIMIT)
    alert('A',"Maximum number of Libraries","have already been opened");
  else {
    if (name[0] == '\0')
      strcpy(defnname,FileList("",DFNEXTN,0));
    else
      strcpy(defnname,name);
    if (strcmp(defnname,"?") == 0)
      alert('A',"There are no Definition files to open !","");
    else if (strlen(defnname) > 0) {
      strcpy(sourceName[maxLibraries],defnname);
      strcat(defnname,DFNEXTN);
      if (!OpenFile(defnname,"r\0",0,' ')) {
        alert('A',"Definition file failed to open",defnname);
        name[0] = '\0';
        }
      else if (!feof(stream)) {
           /* DEFN file becomes a library */
        Defn = AllocDefn(-1);    /* add to tail */
        strcpy(Defn->defn,sourceName[maxLibraries]);
        maxLibraries++;

        defnLineNumber = 0;
        do {    /* read next DEFN record and interpret it */
          fgets(Aline,MAXDEFNLINE,stream);
          if (!feof(stream)) {
            Aline[strlen(Aline)-1] = '\0';
            strRemoveDspace(Aline);
            semi = strpos(Aline,';');
            if (semi == -1)
              alert('A',"invalid definition line header - MISSING a ;",Aline);
            else {
              error = defnBeheader(Aline,semi,&RTnumb,(name[0] == '\0'));  /* strip header */
              if (error == 0) {
                Alineptr = Aline;
                do {   /* interpret acronym sections, one per loop */
                  Alineptr += ++semi;            /* move to next section */
                  semi = strpos(Alineptr,';');
                  if (semi == -1) I = strlen(Alineptr); else I = semi;
                  strncpy(Asection,Alineptr,I);  /* extract acronym section */
                  Asection[I] = '\0';
                     /* interpret acronym section and save in data sets */
                  error = acroCracker(Asection,RTnumb);
                  }
                while ((error == 0) && (semi > -1));
              }
              if (error == 999) {   /* 'END DEFN' -> start next RT */
                error = 0;
                defnLineNumber = 0;
              }
            }
          }
        } while (!feof(stream));
        fclose(stream); stream = NULL;
      }
    }
  }
  free(Aline);
  free(Asection);
}  /* end GetDefn */


void DeleteDefn(void)
     /* delete a DEFN file from disk */
{    char defnname[PATHSIZE];
       /* let user select DEFN filename from list of DFNEXTN files */
    strcpy(defnname,FileList("",DFNEXTN,0));
    if (strcmp(defnname,"?") == 0)
      alert('A',"There are no Definition files to delete !","");
    else if (strlen(defnname) > 0) {  /* user selected a filename */
      strcat(defnname,DFNEXTN);
      if (alertYesNo("Do you actually want to delete the definition file",defnname))
        if (remove(defnname) != 0)
          alert('A',"Unable to remove file : locked or protected ?",defnname);
    }
}  /* end DeleteDefn */


void DiscardAll(void)
     /* clear all sets ready to load/write new definitions */
{
  int R,A;      /* local indexes */
  if ((DefnRoot != NULL) && (!saved)) {
    if (alertYesNo("Do you really wish to discard your amendments ?","Yes/No")) {
         /* release RAM */
      Defn = DefnRoot;
      if (Defn->next != NULL)
        for (Defn = Defn->next; Defn->next != NULL; Defn->next) free(Defn->prev);
      free(Defn);

      for (R = 0; R < maxRecType; R++)  {
        RT = RT_ptr[R];
        free(RT);
      }
      for (A = 0; A < maxAcronym; A++) {
        Acro = Acronym_ptr[A];
        free(Acro);
      }
         /* reset counters and pointers */
      for (A = 0; A < 3; A++) {
        active[A] = 0;
        selected[A] = 0;
      }
      NullDRA();
    }
  }
}  /* end DiscardAll */


int NewRow(int list, int row)
    /* determines the row (= index into one of the DRA sets) number
       <row>  is the requested row
       <list> is one of the DRA lists */
{
    row += Lt[list].start;                    /* add list_start index */
    if (row ==-1) {                           /* if step off top, move to bottom */
      if (Lt[list].used < Lt[list].page) row = Lt[list].used;
      else row = Lt[list].used-1;
      }
    else if (row < -1) row = 0;               /* if jump off top, keep at top */
/*    else if (row == Lt[list].used+1) row = 0; /* if step off bottom, move to top */
    else if (row > Lt[list].used) row = Lt[list].used; /* if off bottom, keep at bottom */

    if (row == Lt[list].start + Lt[list].page)
        Lt[list].start++;
    else if ((row < Lt[list].start) || (row-Lt[list].page > Lt[list].start))
        Lt[list].start = row;
    return row;
}  /* end NewRow */


int DefnRelist(int row, int column)
     /* re-display the definition list
        <row>,<column> = the row and column of the DEFNLIST to make active */
{ int  I;
  int Refresh;  /* true if definition selection has changed */
  int list = DEFNLIST;
  ListRestart(list, column);
  Refresh = ((row < 0) || (row >= Lt[list].page));
  if (Refresh) {
    row = NewRow(list,row);         /* requested row as an index */
    LocateActiveDefn(Lt[list].start);
    Lt[list].displayed = 0;
    for (I = 0; I < Lt[list].page; I++) {
      if (DefnActive == NULL)
        ListDefaultFill(list,0,2,I,0); /* fill the empty list records */
      else {                           /* place selection in list */
        ListFill(list,0,I,DefnActive->defn,0);
        ListFill(list,1,I,"",DefnActive->source);
        ListFill(list,2,I,DefnActive->description,0);
        Lt[list].displayed = I+1;
        LocateActiveDefn(I+1+Lt[list].start);
      }
    }
   }
  else if (row+Lt[list].start > Lt[list].used) row = Lt[list].used;
  else row += Lt[list].start;

     /* reset active[DEFNLIST] */
  LocateActiveDefn(row);
     /* reset list parameters & ensure list is visible */
  if ((Refresh) || (row != Lt[list].row+Lt[list].start)) {  /* update rt selection to this defn */
    RTRelist(-2,0);
    ListRestart(list, column);
    listIsOn[list] = false;
  }
  Lt[list].column = column;
  if (!listIsOn[list]) ListOn(list,column, 1,11,'D',BLACK,BROWN);
  Lt[list].row = MoveDisplayrow(list, Lt[list].row, row-Lt[list].start, column, Refresh);
  return row;
} /* end DefnRelist */


void DefnListInit(int page)
     /* initiallize the definition list DEFNLIST
        <page> is how many rows of list to display */
{   DEFNPAGE = page;
    Lt[DEFNLIST].page  = DEFNPAGE;
    Lt[DEFNLIST].start = 0;
/*    Lt[DEFNLIST].used  = 0; */
    Lt[DEFNLIST].displayed = 0;
    Lt[DEFNLIST].limit = DEFNLIMIT;
    Lt[DEFNLIST].row   = DEFNLIMIT;
    Lt[DEFNLIST].column= 0;
       /* allocate list then set column titles and data controls */
    ListNew(DEFNLIST,"");
    ListFillInit(DEFNLIST,0,"DEFN",TEXTTYPE,2+PLATFORM,0,LENDEFN,0);
    ListFillInit(DEFNLIST,1,"lib",INTEGERTYPE,0,0,LIBLIMIT,0);
    ListFillInit(DEFNLIST,2,"Description",TEXTTYPE,0,0,LENDESC,0);
}  /* end DefnListInit */


int DefnNew(int source, int row, int column)
    /* allocate and fill a new definition
       <source> is the parent library */
{
    if (row < 0) DefnActive = NULL;
    else LocateActiveDefn(row);
    if (DefnActive == NULL) {  /* creating a new definition */
         /* empty fill */
      DefnActive = AllocDefn(-1);
      DefnActive->source = source;
    }
       /* get values from list and put into the definition */
    row -= Lt[DEFNLIST].start;
    switch (column) {
      case 0 : ListGetString(Lcolm, row, DefnActive->defn); break;
      case 1 : DefnActive->source = ListGetValue(Lcolm, row); break;
      case 2 : ListGetString(Lcolm, row, DefnActive->description);
    }
    return 1;
}  /* end DefnNew */


int RTRelist(int row, int column)
     /* re-display the record_type list
        row,column = the row and column of the RTLIST to make active */
{ int  I;
  int Refresh;  /* true if record_type selection has changed */
  int list = RTLIST;
  ListRestart(list, column);
  Refresh = ((row < 0) || (row >= Lt[list].page));
  if (Refresh) {
    if (DefnActive == NULL) {     /* record_type not from a definition */
      for (I = 0; I < Lt[list].page; I++) ListDefaultFill(list,0,1,I,0);
      Lt[list].displayed = 0;
      Lt[list].used = 0;
      row = 0;
      }
    else {                  /* record_type belongs to a definition */
      Lt[list].used = DefnActive->RTused;
      Lt[list].displayed = 0;
      row = NewRow(list,row);
      for (I = 0; I < Lt[list].page; I++) {
        if (I+Lt[list].start >= DefnActive->RTused)
          ListDefaultFill(list,0,1,I,0); /* fill the empty list records */
        else {                          /* place selection in list */
          RT = RT_ptr[DefnActive->RTid[I+Lt[list].start]];
          ListFill(list,0,I,RT->rt,0);
          ListFill(list,1,I,RT->description,0);
          Lt[list].displayed++;
        }
      }
    }
   }
  else if (row+Lt[list].start > Lt[list].used) row = Lt[list].used;
  else row += Lt[list].start;     /* add list_start index */

     /* reset active[RTLIST] */
  if (row >= Lt[list].used) active[list] = maxRecType;
    else active[list] = DefnActive->RTid[row];
     /* update acronym selection to this rt */
  if ((Refresh) || (row-Lt[list].start != Lt[list].row)) {
    AcroRelist(-2,0);
    ListRestart(list, column);
    listIsOn[list] = false;
  }
     /* reset list parameters & ensure list is visible */
  Lt[list].column = column;
  if (!listIsOn[list]) ListOn(list,column, 10,7,'D',BLACK,GREEN);
  Lt[list].row = MoveDisplayrow(list, Lt[list].row, row-Lt[list].start, column, Refresh);
  return row;
}  /* end RTRelist */


void RTListInit(int page)
    /* allocate and fill a new record_type */
{   RTPAGE = page;
    Lt[RTLIST].page  = RTPAGE;
    if (focused)
      Lt[RTLIST].limit = QTYALLOWED;
    else
      Lt[RTLIST].limit = RTLIMIT;
    Lt[RTLIST].start = 0;
    Lt[RTLIST].row   = RTPAGE + 2;  /* force refresh */
    Lt[RTLIST].column= 0;
       /* allocate list then set column titles and data controls */
    ListNew(RTLIST,"");
    ListFillInit(RTLIST,0,"RT",TEXTTYPE,3,0,LENRT,0);
    ListFillInit(RTLIST,1,"Description",TEXTTYPE,0,0,LENDESC,0);
}  /* end RTListInit */


int RTnew(int row, int column)
    /* allocate and fill a new record_type */
{   int R;
    R = maxRecType;
    if (R >= RTLIMIT) {
      alert('A',"Maximum number of Record Types","has already been opened");
      return 0; }
    else  {                           /* update source */
      if (DefnActive != NULL) {       /* record_type belongs to a defn */
        if (row < DefnActive->RTused) /* updating an existing record_type */
          R = DefnActive->RTid[row];
        else if (DefnActive->RTused >= QTYALLOWED) {
          alert('A',"Maximum number of record_types for a definition",
                    "has already been opened");
          return 0; }
        else {                     /* creating a new record_type */
          RT_ptr[R] = (struct RECTYPE *) malloc(sizeof(*RT));
          RT = RT_ptr[R];
          CheckRAM(RT);
          maxRecType++;
             /* empty fill */
          RT->source = activeLibrary;
          RT->rt[0] = '\0';
          RT->description[0] = '\0';
          RT->Aused = 0;
          RT->Aid[0] = 0;
             /* update source */
          DefnActive->RTid[DefnActive->RTused] = R;
          DefnActive->RTused++;
        }
      }
      RT = RT_ptr[R];
         /* get values from list and put into the definition */
      row -= Lt[RTLIST].start;
      switch (column) {
        case 0 : ListGetString(Lcolm, row, RT->rt); break;
        case 1 : ListGetString(Lcolm, row, RT->description);
      }
      active[RTLIST] = R;
      return 1;
    }
}  /* end RTnew */


int AcroRelist(int row, int column)
     /* re-display the acronym list
        row,column = the row and column of the ACROLIST to make active */
{ int  I;
  int Refresh;  /* true if acronym selection has changed */
  int list = ACROLIST;
  ListRestart(list, column);
  Refresh = ((row < 0) || (row >= Lt[list].page) || !(Lcolm->visible));
  if (Refresh) {
    RT = RT_ptr[active[RTLIST]];
    Lrecd = listSet[list];
    if (Lrecd != NULL) {               /* only clear if exits */
      for (I = 0; I <= Lrecd->topColumn; I++) {
        Lcolm = Lrecd->List[I];
        ListReSetColumnInit();
      }
    }
    if (RT == NULL) {       /* acronym not from a record_type */
      for (I = 0; I < Lt[list].page; I++) ListDefaultFill(list,0,5,I,1);
      Lt[list].used = 0;
      Lt[list].displayed = 0;
      row = 0;
      }
    else {                  /* acronym belongs to a record_type */
      Lt[list].used = RT->Aused;
      row = NewRow(list,row);
      Lt[list].displayed = 0;
      for (I = 0; I < Lt[list].page; I++) {
        if (I+Lt[list].start >= RT->Aused)
          ListDefaultFill(list,0,5,I,1); /* fill the empty list records */
        else {                             /* place selection in list */
          Acro = Acronym_ptr[RT->Aid[I+Lt[list].start]];
          ListFill(list,0,I,Acro->acronym,0);
          ListFill(list,1,I,Acro->format,0);
          ListFill(list,2,I,"",Acro->null);
          ListFill(list,3,I,Acro->unit,0);
          ListFill(list,4,I,Acro->name,0);
          ListFill(list,5,I,Acro->comment,0);
          Lt[list].displayed++;
        }
      }
    }
    ListOn(list,column, 15,3,'D',BLACK,CYAN); /* display new list */
   }
  else if (row+Lt[list].start> Lt[list].used) row = Lt[list].used;
  else row += Lt[list].start;     /* add list_start index */
     /* reset list parameters & move to active field */
  Lt[list].column = column;
  Lt[list].row = MoveDisplayrow(list, Lt[list].row, row-Lt[list].start, column, Refresh);
  if (row >= RT->Aused) active[list] = maxAcronym;
  else active[list] = RT->Aid[row];
  return row;
}  /* end AcroRelist */


void AcroListInit(int page)
    /* allocate and fill a new acronym */
{   ACROPAGE = page;
    Lt[ACROLIST].page  = ACROPAGE;
    if (focused)
      Lt[ACROLIST].limit = QTYALLOWED;
    else
      Lt[ACROLIST].limit = ACROLIMIT;
    Lt[ACROLIST].start = 0;
    Lt[ACROLIST].row   = ACROPAGE + 2;  /* force refresh */
    Lt[ACROLIST].column= 0;
       /* allocate list then set column titles and data controls */
    ListNew(ACROLIST,"");
    ListFillInit(ACROLIST,0,"Acronym",TEXTTYPE,1,0,LENACRO,0);
    ListFillInit(ACROLIST,1,"Format",TEXTTYPE,19,2,LENFORMAT,0);
/*    ListFillInit(ACROLIST,2,"Start",INTEGERTYPE,0,1,32700,0); */
    ListFillInit(ACROLIST,2,"Null ",DOUBLETYPE,0,-1000000000.0,1000000000.0,6);
    ListFillInit(ACROLIST,3,"Units",TEXTTYPE,32,0,LENUNIT,0);
    ListFillInit(ACROLIST,4,"Name          ",TEXTTYPE,32,0,LENNAME,0);
    ListFillInit(ACROLIST,5,"Comments",TEXTTYPE,0,0,LENCOMMENT,0);
}  /* end AcroListInit */


int AcroNew(int source,int row, int column)
    /* allocate and fill a new acronym
       <source> = the RT of the acronym
       <row> and <column> are the acronym and parameter indexes */
{   int A;
    A = maxAcronym;
    if (A == ACROLIMIT) {
      alert('A',"Maximum number of Acronyms","has already been opened");
      return 0; }
    else  {                        /* update source */
      RT = RT_ptr[source];
         /* fix up "null" record_type */
      if (RT == NULL) RTnew(row, -1);
         /* find/allocate the acronym */
      if (row < RT->Aused)       /* updating an existing acronym */
        A = RT->Aid[row];
      else if (RT->Aused >= QTYALLOWED) {
        alert('A',"Maximum number of Acronyms for a record type",
                  "has already been opened");
        return 0; }
      else {                     /* creating a new acronym */
        RT->Aid[RT->Aused] = A;
        RT->Aused++;
        Acronym_ptr[A] = (struct ACRONYM *) malloc(sizeof(*Acro));
        Acro = Acronym_ptr[A];
        CheckRAM(Acro);
        maxAcronym++;
           /* empty fill */
        Acro->source = activeLibrary;
        Acro->acronym[0] = '\0';
        Acro->format[0] = '\0';
        Acro->name[0] = '\0';
        Acro->unit[0] = '\0';
        Acro->comment[0] = '\0';
/*          Acro->start = 1; */
        Acro->null = NOVALUE;
      }
      Acro = Acronym_ptr[A];
         /* get values from list and put into the definition */
      row -= Lt[ACROLIST].start;
      switch (column) {
        case 0 : ListGetString(Lcolm, row, Acro->acronym); break;
        case 1 : ListGetString(Lcolm, row, Acro->format); break;
/*        case 2 : Acro->start = ListGetValue(Lcolm, row); break; */
        case 2 : Acro->null  = ListGetValue(Lcolm, row); break;
        case 3 : ListGetString(Lcolm, row, Acro->unit); break;
        case 4 : ListGetString(Lcolm, row, Acro->name); break;
        case 5 : ListGetString(Lcolm, row, Acro->comment);
      }
      active[ACROLIST] = A;
      return 1;
    }
} /* end AcroNew */


void CheckDecimals(char *format, char *addf)
     /* fixes decimal-type C-formats */
{
    if (strchr(format,'.') == NULL) strcat(format,".0");
    strcat(format,addf);
}  /* end CheckDecimals */

void FormatWidths(void)
     /* determine the width of each format in the active definition */
{   int A,R;
    int dataWidth;        /* definition record size */

       /* scan all associated record_types */
    DefnActive->RTnull = -1;                    /* default is RT present */
    for (R = 0; R < DefnActive->RTused; R++) {
      RT = RT_ptr[DefnActive->RTid[R]];
      if (strcmp(RT->rt,"<>"))
        dataWidth = 4;
      else {                                    /* data has no RT in front */
        dataWidth = 0;
        DefnActive->RTnull = R;
      }
         /* scan all associated acronyms of the record_type */
      for (A = 0; A < RT->Aused; A++) {
        Acro = Acronym_ptr[RT->Aid[A]];
        Acro->fwidth = FORTRANwidth(Acro->format);
        Acro->fstart = dataWidth;
        strcpy(Acro->Cformat,Acro->format);  /* convert format from FORTRAN to C */
          Acro->Cformat[0] = '%';
          switch (Acro->format[0]) {
            case 'I' : ;
            case 'L' : strcat(Acro->Cformat,"ld"); break;
            case 'F' : CheckDecimals(Acro->Cformat,"lf"); break;
            case 'E' : ;
            case 'D' : ;
            case 'G' : CheckDecimals(Acro->Cformat,"le"); break;
            case 'X' : ;
            case 'A' : ;
            default  : strcat(Acro->Cformat,"s");
          }
        dataWidth   += Acro->fwidth;
        Acro->order  = strlen(Acro->acronym) + 1;
        if (Acro->order <= Acro->fwidth ) Acro->order = Acro->fwidth + 1;
      }
      RT->order = dataWidth + 1;   /* store data record size, +1 for \n */
    }
}  /* end FormatWidths */


int MoveDRAlistRow(int list, int row)
     /* move the active row of DRA list */
{
    row -= Lt[list].start;
    if (list == DEFNLIST)
      row = DefnRelist(row,Lt[list].column);
    else if (list == RTLIST)
      row = RTRelist(row,Lt[list].column);
    else if (list == ACROLIST)
      row = AcroRelist(row,Lt[list].column);

    return row;
}  /* end MoveDRAlistRow */


int MoveDRAlistColumn(int list, int column)
    /* move the active column of DRA list */
{      /* wrap column request */
    Listline(Lt[list].row,Lcolt,false);
    if      (column < 0) column = Lrecd->topColumn;
    else if (column > Lrecd->topColumn) column = 0;
       /* re-display list */
    if (list == DEFNLIST)
      DefnRelist(Lt[list].row,column);
    else if (list == RTLIST)
      RTRelist(Lt[list].row,column);
    else if (list == ACROLIST)
      AcroRelist(Lt[list].row,column);

    return column;
} /* end MoveDRAlistColumn */


int UpdateDRA(int list, int row, int column)
     /* update the information in the DRA list */
{   int i;
    if (list == DEFNLIST)
      i = DefnNew(activeLibrary,row,column);
    else if (list == RTLIST)
      i = RTnew(row,column);
    else if (list == ACROLIST)
      i = AcroNew(active[list-1],row,column);
    return i;
}  /* end UpdateDRA */


void CopyDRA(int list)
     /* keeps a copy of the active DRA entry */
{
    if ((list == DEFNLIST) && (DefnActive != NULL))
      DefnKept = *DefnActive;
    else if ((list == RTLIST) && (active[list] < maxRecType)) {
      RT = RT_ptr[active[list]];
      RTKept = *RT;
      }
    else if ((list == ACROLIST) && (active[list] < maxAcronym)) {
      Acro = Acronym_ptr[active[list]];
      AcroKept = *Acro;
    }
    selected[list] = true;
} /* end CopyDRA */


void PasteDRA(int list)
     /* overwrites the active DRA entry with the keep entry */
{ if (selected[list]) {
    if (list == DEFNLIST) {
      if (DefnActive == NULL) {
        DefnNew(activeLibrary,-1,-1);
        if (Lt[list].displayed < Lt[list].page) Lt[list].displayed++;
      }
      *DefnActive = DefnKept;
      }
    else if (list == RTLIST) {
      if (active[list] >= maxRecType) RTnew(maxRecType,-1);
      RT  = RT_ptr[active[list]];
      *RT = RTKept;
      }
    else if (list == ACROLIST) {
      if (active[list] >= maxAcronym) AcroNew(active[list-1],maxAcronym,-1);
      Acro  = Acronym_ptr[active[list]];
      *Acro = AcroKept;
    }
       /* redisplay the list */
    Lt[list].row = -1;
    MoveDRAlistColumn(list, Lt[list].column);
    saved = false;
  }
} /* end PasteDRA */


void DRAins(int list, int row)
     /* insert a record into a DRA list
        by scrolling pointers up from row
        after ensuring space is available */
{   int I;
    if (list == DEFNLIST) {
      if (DefnActive != NULL) DefnNew(activeLibrary,row,-1);
      }
    else if (list == RTLIST) {
      if ((row < QTYALLOWED)
      && ( row < DefnActive->RTused)
      && ( RTnew(maxRecType,-1))) {
        for (I = DefnActive->RTused; I > row; I--)
          DefnActive->RTid[I] = DefnActive->RTid[I-1];
        DefnActive->RTid[row] = DefnActive->RTid[DefnActive->RTused];
      }
    }
    else if (list == ACROLIST) {
      RT = RT_ptr[active[list-1]];
      if ((row < QTYALLOWED)
      && ( row < RT->Aused)
      && ( AcroNew(active[list-1],maxAcronym,-1))) {
        for (I = RT->Aused; I > row; I--) RT->Aid[I] = RT->Aid[I-1];
        RT->Aid[row] = RT->Aid[RT->Aused];
      }
    }
       /* fixup then redisplay the list */
    ListInsertRow(list, row);
    if (Lt[list].displayed < Lt[list].page) Lt[list].displayed++;
    Lt[list].row = -1;
    MoveDRAlistColumn(list, Lt[list].column);
    saved = false;
}  /* end DRAins */


struct DEFN *DisposeDefn(struct DEFN *DefnA)
     /* disposes of the Active DEFN */
{   struct DEFN *Defn;
    Defn = DefnA->next;
    if (Defn != NULL)
      Defn->prev = DefnA->prev;
    else
      DefnTail = DefnA->prev;

    if (DefnA->prev != NULL)
      (DefnA->prev)->next = Defn;
    else
      DefnRoot = Defn;

    free(DefnA);
    return Defn;
}  /* end DisposeDefn */


void DRAdel(int list, int row)
     /* delete a record from a DRA list
        by scrolling pointers down to row */
{   int I;
    int OK = 1;
    if (list == DEFNLIST) {
      LocateActiveDefn(row);
      if (DefnActive == NULL)
        OK = 0;
      else
        DefnActive = DisposeDefn(DefnActive);
     }
    else if (list == RTLIST) {
      if ((DefnActive != NULL) && (row < DefnActive->RTused)) {
        RT_ptr[DefnActive->RTid[row]]->source = -1;
        for (I = row; I < DefnActive->RTused; I++)
           DefnActive->RTid[I] = DefnActive->RTid[I+1];
        DefnActive->RTused--;
        }
      else OK = 0;
     }
    else if (list == ACROLIST) {
      RT = RT_ptr[active[list-1]];
      if (row < RT->Aused) {
        Acronym_ptr[RT->Aid[row]]->source = -1;
        for (I = row; I < RT->Aused; I++) RT->Aid[I] = RT->Aid[I+1];
        RT->Aused--;
        }
      else OK = 0;
    }
       /* fixup then redisplay the list */
    if (OK) {
      ListDeleteRow(list, row);
      Lt[list].used--;
      Lt[list].row = -1;
      MoveDRAlistColumn(list, Lt[list].column);
      saved = false;
    }
}  /* end DRAdel */


int alertError(int error, char *message, char *second)
    /* displays an alert then returns the given error code */
{   alert('A',message,second);
    return error;
}  /* end alertError */


int defnBeheader(char *Aline, int semi, int *RTnumb, int warnRT)
    /* extracts and interprets a DEFN header */
{   int I,P,R;
    int error = 0;
    int dnumb;          /* DEFN record sequence number */
    char header[100];   /* header section of DEFN record */
    char message[76];   /* message to user */
    char RTname[LENRT]; /* record_type name in header */
    char rtname[LENRT]; /* record_type name in database */
    char *head;         /* pointer into header */
    char *ptr;          /* pointer into header */
    char *ptrS;         /* pointer into header to location of "ST=" */
    struct DEFN *Defn;  /* holds pointer */

       /*  Check if ST & RT exist */
    strncpy(header,Aline,semi);
    header[semi] = '\0';
    strRemoveAllspace(header);
    P = 4;
    while ((header[P] >= '0') && (header[P] <= '9') && (P <= semi)) P++;
    if ((header[0] != 'D') || (header[1] != 'E') || (header[2] != 'F') || (header[3] != 'N'))
      error = alertError(79,"definition does not start with 'DEFN' in",Aline);
    else if (P > 8)
      error = alertError(80,"definition sequence count error - TOO BIG in",Aline);
    else if (P > 4)                 /* read the DEFN sequence number */
      sscanf(header,"DEFN%i",&dnumb);
    else                            /* no sequence number */
      dnumb = 0;

    if (error == 0) {
      defnLineNumber++;
      if (dnumb != defnLineNumber)
        error = alertError(80,"definition sequence count error - OUT OF SEQUENCE in",Aline);
      else {
        head = header + P;
        ptr = strstr(head,"ST=");
        if (ptr == NULL)
          error = alertError(81,"missing Storage_Type in",Aline);
        else {
          ptrS = strstr(head,"ST=RECD");
          if (ptrS == NULL)
            error = alertError(81,"un-recognised Storage_Type in",Aline);
          else {
            ptr = strstr(head,"RT=");
            if (ptr == NULL)
              error = alertError(83,"missing Record_Type in",Aline);
            else {
              I = 3;
              while ((I-2 < LENRT) && (ptr[I] != ',') && (ptr[I] != '\0')
                && (ptr[I] != ';') && (ptr != ptrS))
                RTname[I-3] = ptr[I++];
              RTname[I-3] = '\0';
              if (RTname[0] != '\0') strcpy(rtname,RTname); /* not absent */
              else strcpy(rtname,"<>");                     /* absent */
                 /* Check for bad record type name */
              if (defnLineNumber > 1) {
                *RTnumb = maxRecType - 1;
                if (strcmp(rtname,RT_ptr[*RTnumb]->rt))
                  error = alertError(83,"Record_Type mis-match in",Aline);
                }
              else {
                   /* warn of duplicated record type name */
                if (warnRT) for (R = 0; R < maxRecType; R++) {
                  RT = RT_ptr[R];
                  if (strcmp(RT->rt,rtname) == 0) {
                    sprintf(header,"Warning : duplicated Record_Type name %s",RTname);
                    Defn = DefnActive;             /* store active defn */
                    LocateActiveDefn(RT->source);
                    sprintf(message,"in Library %s",sourceName[DefnActive->source]);
                    DefnActive = Defn;             /* recover active defn */
                    alert('A',header,message);
                    R = maxRecType;                /* warning made - skip rest */
                  }
                }
                   /* must be a new record_type */
                *RTnumb = maxRecType;
                RTnew(maxRecType,-1);
                strcpy(RT_ptr[*RTnumb]->rt,rtname);
              }
            }
          }
        }
      }
    }
    return error;
}  /* end defnBeheader */


int whichAttrib(char *sect, int *In)
    /* determines which acronym attribute is present
       in string <sect>
       starting at the <In>'th character
       <In> returns as the index to the start of the information in string */
{   int P = 0;    /* usually a position index into string */
    int I;        /* holds local value of <In> */
    char test[5]; /* the first characters of string */
    I = *In + 1;  /* skip colon or comma */

    while (sect[I] == ' ') I++;         /* skip leading spaces */
    *In = I+4;                          /* start of data */
    for (P = 0; P < 4; P++) test[P] = toupper(sect[I++]);
    test[4] = '\0';
    if (strcmp(test,"UNIT") == 0) {
      if (toupper(sect[I]) == 'S') *In = I+1;  /* UNITS also allowed */
      return 1;
      }
    else if (strcmp(test,"NULL") == 0) return 2;
    else if (strcmp(test,"NAME") == 0) return 3;
    else { *In = *In-4; return 4; }            /* no keyword */
}  /* end whichAttrib */


int testEqual(char *info, char *source, int I, int cut)
    /* returns information content of 'source' in 'info'
       I   is the starting point in source
       cut is the maximum return length (= 0 if no limit) */
{   int P = 0;    /* usually a position index into string */
    while (source[I] == ' ') I++;   /* skip spaces */
    if ((cut > 0) && (source[I] == '=')) { /* not an acronym comment */
      I++;
      while ((source[I] != '\0') && (source[I] != ','))
        info[P++] = source[I++];
      if (P > cut-1) P = cut-1;
      }
    else                /* an acronym comment */
      do {              /* to next comma */
        while ((source[I] != '\0') && (source[I] != ',')) info[P++] = source[I++];
        }
        /* check if the comma is inside a comment */
      while ((source[I] == ',') && (whichAttrib(source, &I) > 3));

    info[P] = '\0';
    return I;
}  /* end testEqual */


int acroCracker(char *Asection, int RTnumb)
    /* interprets acronym section 'Asection' and saves in data sets
       RTnumb = parent record_type */
{   int I,P,A;          /* local indexes */
    int colon;          /* location of colon in Asection */
    int error = 0;      /* return value */
    int len;            /* length of Asection */
    int Acrostart;      /* "start" value of ASEG-GDF v1 */
    char section[1000]; /* a piece of Asection */

      /*  extract acronym */
    strRemoveLspace(Asection);
    strRemoveRspace(Asection);
    len = strlen(Asection);
    I = 0;
    while ((Asection[I] != ':') && (I < LENACRO-1) && (I < len))
      section[I] = Asection[I++];
    section[I] = '\0';
    strRemoveRspace(section);
    I = strlen(section);
    if (I == 0)
      error = alertError(90,"error - MISSING acronym name in",Asection);
    else {
      if ((strcmp(section,"END DEFN") == 0)
      || ( strcmp(section,"ENDDEFN") == 0))
        error = 999; /* terminator */
      else {
        colon = strpos(Asection,':');
        if (colon == -1)
          error = alertError(87,"error - MISSING : in",Asection);
        else {
            /* skip extra acronym chars, if any */
          if ((Asection[I] != ' ') && (I < colon)) {
            alert('A',"WARNING - acronym name truncated in",Asection);
            while ((Asection[I] != ' ') && (I < colon)) I++;
          }
          RT = RT_ptr[RTnumb];
          P = RT->Aused;
            /* Reject leading 'RT' acronymn */
          if ((P > 0) || (strcmp(section,"RT"))) {
            /* Check for duplicated acronym */
          for (A = 0; A < P; A++) {
            Acro = Acronym_ptr[RT->Aid[A]];
            if (strcmp(Acro->acronym,section) == 0)
              error = alertError(91,"Error - duplicated acronym in",Asection);
          }

          if (error == 0) {
              /* create new instance of acronym */
            AcroNew(RTnumb,maxAcronym,-1);
            strcpy(Acro->acronym,section);
              /*  extract Start-instance number *** but do not save it */
            for (P = I; P < colon; P++) section[P-I] = Asection[P];
            section[P-I] = '\0';
            P = sscanf(section,"%i",&Acrostart);
/*           if ((P == 0) || (P == isEof)) Acro->start = 1; */
              /* extract format */
            I = colon+1;
            Asection[colon] = ' ';
            colon = strpos(Asection,':');
            Asection[I-1] = ':';
            while (Asection[I] == ' ') I++;   /* skip leading spaces */
            P = 0;
            while ((Asection[I] != ':') && (I < len) && (P < LENFORMAT-1))
              Acro->format[P++] = Asection[I++];
            while (Acro->format[P-1] == ' ') P--;  /* drop trailing spaces */
            Acro->format[P] = '\0';
            if ((P < 2) || (Verify(16,Acro->format) != 0))
              error = alertError(87,"error in acronym - BAD format in",Asection);
            else if (colon != -1) {
                /* extract attributes, if any */
              I = colon;
              while (I < len-1) {
                A = whichAttrib(Asection, &I);   /* I returns as start of data */
                switch (A) {
                  case  1 : P = LENUNIT; break;  /* UNIT */
                  case  2 : P = len;     break;  /* NULL */
                  case  3 : P = LENNAME; break;  /* NAME */
                  default : P = 0;               /* comment */
                }
                I = testEqual(section,Asection,I,P); /* section = information content */
                switch (A) {
                  case  1 : strcpy(Acro->unit,section); break;
                  case  2 : sscanf(section,"%lf",&Acro->null); break;
                  case  3 : strcpy(Acro->name,section); break;
                  default : strncat(Acro->comment,section,LENCOMMENT-strlen(Acro->comment));
                            Acro->comment[LENCOMMENT-1] = '\0';
                }
              } /* end extract attributes */
            }
          }}
        }
      }
    }
    return error;
}  /* end acroCracker */

#endif