/* GDFlists
      handles the creation, filling, display, and retrieval from lists
*/

#ifndef LISTSETS

   /* list sizes */
#define LISTSETS     7   /* maximum qty of edit lists */
#define LISTLIMIT  250   /* maximum list entries */
#define LISTLEN    101   /* maximum length of list display text */
#define LISTLENSHOW 64   /* maximum length of shown list display text */
#define LISTCOLUMNS  7   /* maximum number of data columns in a set */

   /* list field types */
#define DOUBLETYPE    8
#define FLOATTYPE     4
#define TEXTTYPE      9
#define INTEGERTYPE   2

   /* Edit entry error messages */
char errMsg[8][18] = { "               "
                      ,"text too short\a","text too long\a"
                      ,"invalid value\a"
                      ,"value too small\a","value too large\a"
                      ,"must be Y or N!\a"
                      ,"invalid format\a" };


int  listIsOn[LISTSETS];   /* true if list is visible */
struct ListColumn {        /* controls position and action for a list column */
    char  columnTitle[76];     /* header at top */
    int   columnWidth;         /* width in characters */
    int   columnLeft;          /* screen column of left side */
    int   qtyInColumn;         /* rows containing data */
    char  Cform;               /* data form {text, float, integer} */
    int   Cverify;             /* data verification/adjustment code */
   double Cmin,Cmax;           /* min & max (numeric) value; (text) length */
    int   Cdeci;               /* number of decimals to display (float) */
    int   visible;             /* true if the column is visible */
    void  *item[LISTLIMIT];    /* the data */
} *Lcolm;

struct Listrec {           /* for each list */
    char listTitle[76];        /* title at top of list */
    int  offset;               /* size of list header in rows
                                  a blank title means 2 fewer rows in offset */
    struct ListColumn *List[LISTCOLUMNS];  /* the columns */
    int  qtyInList;            /* rows containing data */
    int  topColumn;            /* highest numbered column defined */
    int  columnmin,columnmax;  /* minimum & maximum editable columns */
    int  activeColumn;         /* column highlighted */
    int  listLeft,listTop,listRight,listBottom;  /* screen window */
    int  rowOrigin;            /* data index of top row displayed
                                  allows for vertical scrolling */
    int  displayed;            /* rows visible */
    int  Larea[4000];          /* stores image of screen covered by list */
} *Lrecd;
struct Listrec *listSet[LISTSETS];

   /* handlers for text data for different lengths of text - saves RAM */
struct string9 {
    char content[9];
} *rD9;
struct string21 {
    char content[21];
} *rD21;
struct string41 {
    char content[41];
} *rD41;
struct string101 {
    char content[101];
} *rD101;

   /* handlers numeric data */
struct integer2 {
    int  content;
} *rD2;
struct float4 {
    float content;
} *rD4;
struct double8 {
    double content;
} *rD8;


int textInsert = 0;           /* toggle between overwrite=1 & insert=0 modes */
unsigned int outkey;          /* key returned from edit */
char *SortTest, *SortPivot;   /* used in sorting list rows */
int  ListOrigin[LISTLIMIT];   /* original row numbers in ListEdit style 4 */
int  cutItem = LISTLIMIT-1;   /* index to a cut item in style 4 */
int  cutOrigin;               /* holds cutItem original row */
char filename[PATHSIZE];      /* holds a user-selected filename */

   /* alterable colour options */
int Lcolt  = WHITE;   /* Text colour */
int Lcolb  = BLUE;    /* Background colour */
int Lcolab = BLUE;    /* Current choice active colour */
int Lcolat = YELLOW;  /* Current choice text colour */
int Lcoledit = GREEN; /* Colour of edit text */


/* prototypes */


void NullLists(void);
     /* allocate heap RAM to list pointers */
void RowDispose(struct ListColumn *Lcolm, int row);
     /* release RAM used by list column <Lcolm> <row> */
void ListColumnsDispose(int list);
     /* release RAM used by <list> columns */
void ListNew(int list, char *header);
     /* creates a new <list> with title of <header> */
void FillListRow(struct ListColumn *Lcolm, int row, char *description, double value);
     /* puts data into list column <Lcolm> on <row> */
void ListInsertRow(int list, int row);
     /* scroll and insert default row into <list> at <row> */
void ListDeleteRow(int list, int row);
     /* scroll to remove <list> at <row> */
void ListNewRow(struct ListColumn *Lcolm, int row, char *description, double value);
     /* after allocating RAM puts data into list column <Lcolm> on <row> */
void ListitemDuplicate(struct ListColumn *Lcolm, int source, int destination);
     /* copy a list entry of column <Lcolm>
        from row <source> to unallocated row <destination> */
void ListFillInit(int list, int column, char *title, char itemtype,
       int verify, double vmin, double vmax, int deci);
     /* fill the control values of a <column> of <list> */
void ListFill(int list, int column, int row, char *description, double value);
     /* put data (<description> or <value>) into <list> <column> <row> */
void ListDefaultFill(int list, int cols, int colf, int row, int deflt);
     /* put default data into <list> <cols..colf> <row> */
int  ListFixFill(int list, int row);
     /* if no entry in <list> column 0 on <row>, place a '*' there */
void SetColumnMin(int list, int min);
     /* set the <min>imum column of <list> that the user can access */
void SetColumnMax(int list, int max);
     /* set the <max>imum column of <list> that the user can access */
void ListOff(int list);
     /* remove <list> from view */
void ListWindow(int column);
     /* sets the text window to that of <column> for the active list */
void Listline(int row, int colour, int highlight);
     /* displays a <row> of the active column of the active list
        text in <colour>
        and if <highlight> then change colour of first character */
void ListaColumn(int column);
     /* display all records of <column> if column visible */
void ListDisplayValues(void);
     /* display all fields of active list */
int ListLR(int L, int activeCol);
    /* determine the screen position of the columns of a list
       hide columns that lie off screen
       <L> is the left edge of the list in screen columns
       return the right edge of list */
void ListRefill(int list, int origin, int display);
     /* redraws <list> from row <origin> with page size <display> */
void ListOn(int list, int colm, int left,int top, char frame, int Tcolour, int Bcolour);
     /* display <list> with screen position <left>,<top>
        using <frame> type {single,double}
        <Tcolour> & <Bcolour> are text and background colours */
void ListRedisplay(int list, int row, int column);
     /* redisplay <list> making <column> and <row> active */
int  MoveDisplayrow(int list, int atrow, int row, int column, int Refresh);
    /* changes the active row of <list> from <atrow> to <row> of <column> */
int  MoveAtListrow(int row, int column, int newrow, int Refresh);
    /* changes row of the active list from <row> to <newrow> of <column>
       <Refresh> true forces a rewrite of all fields */
int  MoveAtListcolumn(int list, int row, int column, int newcolumn);
    /* changes column of <list> from <column> to <newcolumn> */
void ListRestart(int list, int column);
     /* make <list> active on <column> */
void StoreCut(int row);
     /* makes a copy of a list slot */
void PasteCut(int row);
     /* places cut item in a list slot */
void DelCut(int list, int column, int row);
     /* deletes a row from a single column */
void InsCut(int list, int column, int row);
     /* insert a default row into a single column */
void ShowChar(int where, char thechar, int colour, int keep);
     /* write character <thechar> in active field
        at field column <where> in <colour> */
void renew(char *Ttemp, int startchar, int limit);
     /* rewrites a field with text <Ttemp>
        from the <startchar> char of <Ttemp>
        for no more than <limit> chars */
void straddchr(char *Ttemp, char editkey, int used);
     /* adds char <editkey> to end of <Ttemp> */
void ListEditField(unsigned char Ttype, double Tmin, double Tmax, int verify,
               unsigned char editkey, char *usertext);
     /* edits the displayed field entry of active list column row
        <Ttype>    is the form
        <Tmin>,<Tmax> are the min & max limits for the column data
        <verify>   is the verification code
        <editkey>  is the key that the user pressed to initiate the edit
                   input >127 to keep initial usertext
        <usertext> is the initial and final field entry as text */
unsigned int DoEdit(int row, int column, unsigned char option, int list);
     /* sets up a field for editing and restores list edit on return */
int  ListEdit(int list, int left, int top, int column, int row, int style);
    /* general list editor without regard to the source of the data */
void SortLR(int l, int r);
     /* a QuickSort of list entries in a list column */
void SortList(int l, int r);
     /* padding needed by C for sorting */
int  GetFileList(int list, char *title, char *fextn);
     /* create a list of files found on the system */
void FileListChoice(int list, int start, char *title, char *fileExn);
     /* get and display a list of files, allowing the user to pick one */
char *FileList(char *title, char *extn, int start);
     /* let the user select from a list of files matching <extn> */

int  ListGetQty(int list);
     /* returns the rows allocated to a list */
void TrimZero(char *tempstr);
     /* removes trailing zeros (& dot if on end) of <tempstr> */
void ListGetString(struct ListColumn *Lcolm, int row, char *string);
     /* returns the string value of a list field */
double ListGetValue(struct ListColumn *Lcolm, int row);
     /* returns the numerical value of a list field */
int  ListGetOriginal(int row);
     /* returns the original row number before editing moved row values */


/* implimentation */


void NullLists(void)
     /* allocate heap RAM to list pointers */
{   int i,column;
    for (i = 0; i < LISTSETS ; i++) {
      listSet[i] = (struct Listrec *) malloc(sizeof(*Lrecd));
      Lrecd = listSet[i];
      CheckRAM(Lrecd);                         /* verify RAM allocated */
      for (column = 0; column < LISTCOLUMNS; column++) {
        Lrecd->List[column] = NULL;
        Lrecd->topColumn = -1;                 /* first column = 0 */
      }
    }
}  /* end NullLists */

void RowDispose(struct ListColumn *Lcolm, int row)
     /* release RAM used by list column <Lcolm> <row> */
{
    if (Lcolm->item[row] != NULL) {
          switch (Lcolm->Cform) {
/* integer */ case  2 : rD2 = (struct integer2 *) Lcolm->item[row];
                     free(rD2); break;
/* float */   case  4 : rD4 = (struct float4 *) Lcolm->item[row];
                     free(rD4); break;
/* double*/   case  8 : rD8 = (struct double8 *) Lcolm->item[row];
                     free(rD8); break;
/* text9 */   case  9 : rD9 = (struct string9 *) Lcolm->item[row];
                     free(rD9); break;
/* text21 */  case 21 : rD21 = (struct string21 *) Lcolm->item[row];
                     free(rD21); break;
/* text41 */  case 41 : rD41 = (struct string41 *) Lcolm->item[row];
                     free(rD41); break;
/* text101*/  case 101: rD101 = (struct string101 *) Lcolm->item[row];
                     free(rD101);
          }
    }
}  /* end RowDispose */


void ListColumnsDispose(int list)
     /* release RAM used by <list> columns */
{
    int column,row;
    ListOff(list);                /* make sure display cleared */
       /* release each column */
    for (column = 0; column <= Lrecd->topColumn; column++) {
      Lcolm = Lrecd->List[column];
           /* release each row of data */
      if (Lcolm != NULL) {
        for (row = 0; row < Lcolm->qtyInColumn; row++) RowDispose(Lcolm,row);
        free(Lrecd->List[column]);
        Lrecd->List[column] = NULL;
      }
    }
} /* end ListColumnsDispose */


void ListNew(int list, char *header)
     /* creates a new <list> with title of <header> */
{
    ListColumnsDispose(list);        /* release any RAM from previous use */
       /* set title and an empty list */
    strcpy(Lrecd->listTitle,header);
    Lrecd->qtyInList  =  0;
    Lrecd->topColumn  = -1;          /* first column = 0 */
    Lrecd->rowOrigin  =  0;
    Lrecd->columnmin  =  0;
    Lrecd->columnmax  =  LISTCOLUMNS;
    listIsOn[list]    =  false;
} /* end ListNew */


void FillListRow(struct ListColumn *Lcolm, int row, char *description, double value)
     /* puts data into list column <Lcolm> on <row>
        if a text column then data = <description>
        else if <description> != NULL then data = value of <description>
        else data = <value> */
{
    switch (Lcolm->Cform) {
/* numeric columns */
      case  2 : rD2 = (struct integer2 *) Lcolm->item[row];
                if (strlen(description) == 0) rD2->content = value;
                else rD2->content = atoi(description); break;
      case  4 : rD4 = (struct float4 *) Lcolm->item[row];
                if (strlen(description) == 0) rD4->content = value;
                else rD4->content = atof(description); break;
      case  8 : rD8 = (struct double8 *) Lcolm->item[row];
                if (strlen(description) == 0) rD8->content = value;
                else rD8->content = atof(description); break;
/* text columns */
      case  9 : rD9 = (struct string9 *) Lcolm->item[row];
                strcpy(rD9->content,description); break;
      case 21 : rD21 = (struct string21 *) Lcolm->item[row];
                strcpy(rD21->content,description); break;
      case 41 : rD41 = (struct string41 *) Lcolm->item[row];
                strcpy(rD41->content,description); break;
      case 101: rD101 = (struct string101 *) Lcolm->item[row];
                strcpy(rD101->content,description);
    }
} /* end FillListRow */


void ListInsertRow(int list, int row)
     /* scroll and insert default row into <list> at <row> */
{   int r,c;
    void *ptr;
    Lrecd = listSet[list];
    for (c = 0; c <= Lrecd->topColumn; c++) {
      Lcolm = Lrecd->List[c];
      ptr = Lcolm->item[Lrecd->displayed-1];
      for (r = Lrecd->displayed-1; r > row - Lrecd->rowOrigin; r--)
        Lcolm->item[r] = Lcolm->item[r-1];
      Lcolm->item[row] = ptr;
    }
    ListDefaultFill(list, 0,Lrecd->topColumn, row, 0);
} /* end ListInsertRow */


void ListDeleteRow(int list, int row)
     /* scroll to remove <list> at <row> */
{   int r,c;
    void *ptr;
    int d;
    Lrecd = listSet[list];
    d = Lrecd->displayed-1;
    for (c = 0; c <= Lrecd->topColumn; c++) {
      Lcolm = Lrecd->List[c];
      ptr = Lcolm->item[row - Lrecd->rowOrigin];
      for (r = row - Lrecd->rowOrigin; r < d; r++)
        Lcolm->item[r] = Lcolm->item[r+1];
      Lcolm->item[d] = ptr;
    }
    ListDefaultFill(list, 0,Lrecd->topColumn, d + Lrecd->rowOrigin, 0);
} /* end ListDeleteRow */


void ListNewRow(struct ListColumn *Lcolm, int row, char *description, double value)
     /* after allocating RAM
        put data (<description> or <value>) into list column <Lcolm> on <row> */
{
    switch (Lcolm->Cform) {
/* numeric columns */
      case  2 : rD2 = (struct integer2 *) malloc(sizeof(*rD2));
                Lcolm->item[row] = rD2; break;
      case  4 : rD4 = (struct float4 *) malloc(sizeof(*rD4));
                Lcolm->item[row] = rD4; break;
      case  8 : rD8 = (struct double8 *) malloc(sizeof(*rD8));
                Lcolm->item[row] = rD8; break;
/* text columns */
      case  9 : rD9 = (struct string9 *) malloc(sizeof(*rD9));
                Lcolm->item[row] = rD9; break;
      case 21 : rD21 = (struct string21 *) malloc(sizeof(*rD21));
                Lcolm->item[row] = rD21; break;
      case 41 : rD41 = (struct string41 *) malloc(sizeof(*rD41));
                Lcolm->item[row] = rD41; break;
      case 101: rD101 = (struct string101 *) malloc(sizeof(*rD101));
                Lcolm->item[row] = rD101;
    }
       /* put data into column row */
    FillListRow(Lcolm,row,description,value);
} /* end ListNewRow */


void ListitemDuplicate(struct ListColumn *Lcolm, int source, int destination)
     /* copy a list entry of column <Lcolm>
        from row <source> to unallocated row <destination> */
{   char description[101];
    double value;

    if (Lcolm->Cform > 8) {  /* text */
      ListGetString(Lcolm, source, description);
      value = 0.0;
      }
    else {                   /* numeric */
      description[0] = '\0';
      value = ListGetValue(Lcolm, source);
    }
    ListNewRow(Lcolm, destination, description, value);
} /* end ListitemDuplicate */

void ListReSetColumnInit(void)
{
    Lcolm->qtyInColumn = 0;
    Lcolm->columnLeft  = 1;
         /* column must be as wide as title and at least 4 wide */
    Lcolm->columnWidth = maxInteger(4,strlen(Lcolm->columnTitle)+1);
}
void ListFillInit(int list, int column, char *title, char itemtype,
       int verify, double vmin, double vmax, int deci)
     /* fill the control values of a <column> of <list>
        if necessary, create the column.
        <title>    is the <column> header
        <itemtype> is the data's form {text, float, integer}
        <verify>   is the data verification/adjustment code
        <vmin>,<vmax> are the min & max value (numeric type)
                              min & max length (text type)
        <deci>     is the number of decimals to display (float type only) */
{
    int len,row;
    Lrecd = listSet[list];
    if (Lrecd->List[column] == NULL) {    /* need to create column */
      Lrecd->List[column] = (struct ListColumn *) malloc(sizeof(*Lcolm));  /* allocate RAM */
      Lcolm = Lrecd->List[column];
      CheckRAM(Lcolm);                    /* verify RAM allocated */
      for (row= 0; row < LISTLIMIT; row++) Lcolm->item[row] = NULL;
      }
    else {                                /* clear column */
      Lcolm = Lrecd->List[column];
      for (row= 0; row < Lcolm->qtyInColumn; row++) RowDispose(Lcolm,row);
    }
       /* fill column controls */
    strcpy(Lcolm->columnTitle,title);
    ListReSetColumnInit();
    if (column > Lrecd->topColumn) Lrecd->topColumn = column;
    Lcolm->Cmin  = vmin;
    Lcolm->Cmax  = vmax;
    Lcolm->Cdeci = deci;
    Lcolm->Cverify = verify;
       /* set data content form */
    if (itemtype >= TEXTTYPE) {
      if (Lcolm->columnWidth < vmin) Lcolm->columnWidth = vmin;
      len = vmax;
      if      (len < 10) Lcolm->Cform =  9;
      else if (len < 22) Lcolm->Cform = 21;
      else if (len < 42) Lcolm->Cform = 41;
      else Lcolm->Cform = 101;
      }
    else if (itemtype == INTEGERTYPE) Lcolm->Cform = 2;
    else if (itemtype == FLOATTYPE) Lcolm->Cform = 4;
    else Lcolm->Cform = 8;  /* default = double type */
       /* make long entries scroll horizontally */
    if (Lcolm->columnWidth > LISTLENSHOW) Lcolm->columnWidth = LISTLENSHOW;
} /* end ListFillInit */


void ListFill(int list, int column, int row, char *description, double value)
     /* put data (<description> or <value>) into <list> <column> <row> */
{   int I;
    char nullstr[20];
    Lrecd = listSet[list];
    if (Lrecd != NULL) {               /* only fill if exits */
      Lcolm = Lrecd->List[column];
      if (row < Lcolm->qtyInColumn)   /* overwriting current data */
        FillListRow(Lcolm,row,description,value);
      else {                           /* adding a row */
        for (I = Lcolm->qtyInColumn; I <= row; I++)  /* fill down to row */
          ListNewRow(Lcolm,I, description,value);
        Lcolm->qtyInColumn = row+1;
        Lrecd->qtyInList = maxInteger(Lrecd->qtyInList,Lcolm->qtyInColumn);
      }
         /* increase column width if new data longer then current width */
      if (Lcolm->Cform > 8)     /* text column */
        Lcolm->columnWidth = maxInteger(Lcolm->columnWidth,strlen(description)+1);
      else if (value != 0) {     /* numeric column */
        sprintf(nullstr,"%.*f",6,value);
        TrimZero(nullstr);
        Lcolm->columnWidth = maxInteger(Lcolm->columnWidth,strlen(nullstr));
       }
         /* make long entries scroll horizontally */
      Lcolm->columnWidth = minInteger(Lcolm->columnWidth,LISTLENSHOW);
    }
} /* end ListFill */


void ListDefaultFill(int list, int cols, int colf, int row, int deflt)
     /* put default data into <list> <cols..colf> <row> */
{   int C;
    for  (C = cols; C <= colf; C++) {
      if (listSet[list]->List[C]->Cform == 8)
        ListFill(list,C,row,"",NOVALUE);
      else
        ListFill(list,C,row,"",deflt);
    }
    ListOrigin[row] = -1;               /* no source for new row */
}  /* end ListDefaultFill */


int ListFixFill(int list, int row)
     /* if no entry in <list> column 0 on <row>, put in a default
        return flag=1 if changed */
{   char temp[9];
    ListGetString(listSet[list]->List[0],row,temp);  /* get column 0 entry */
    if (strcmp(temp,""))                /* not empty */
      return 0;
    else {                              /* empty */
      if (list != DEFNLIST)
        ListFill(list,0,row,"<>",0.0);  /* put "<>" in as default, for RT means none */
      else {
        sprintf(temp,"%d",row);         /* for DEFN use row number; need a valid DOS name */
        ListFill(list,0,row,temp,0.0);
      }
      ListRedisplay(list, row, 0);
      return 1;
    }
}  /* end ListFixFill */


void SetColumnMin(int list, int min)
     /* set the <min>imum column of <list> that the user can access */
{
    Lrecd = listSet[list];
    Lrecd->columnmin = min;
}


void SetColumnMax(int list, int max)
     /* set the <max>imum column of <list> that the user can access */
{
    Lrecd = listSet[list];
    Lrecd->columnmax = max;
}


void ListOff(int list)
     /* remove <list> from view */
{
    Lrecd = listSet[list];
    if (listSet[list] != NULL) {  /* if exists */
      if (listIsOn[list]) {       /* if visible */
        puttext(Lrecd->listLeft,Lrecd->listTop,
                minInteger(screenRight,Lrecd->listRight),Lrecd->listBottom,Lrecd->Larea);
        listIsOn[list] = false;
      }
    }
}  /* end ListOff */


void ListWindow(int column)
     /* sets the text window to that of <column> for the active list */
{
    Lcolm = Lrecd->List[column];
    if (Lcolm->visible)
      window(Lcolm->columnLeft,Lrecd->listTop+Lrecd->offset,
             Lcolm->columnLeft+Lcolm->columnWidth-1,Lrecd->listBottom);
}  /* end ListWindow */


void Listline(int row, int colour, int highlight)
     /* displays a <row> of the active column of the active list
        text in <colour>
        and if <highlight> then change colour of first character */
{   char value[PATHSIZE];
    int  N,line;
    gotoxy(1,row-Lrecd->rowOrigin+1); clreol(); /* erase any display in field */
    if (row < Lcolm->qtyInColumn) {             /* data exists */
      textcolor(colour);
      ListGetString(Lcolm,row,value);
      if (highlight) {                          /* highlight first char */
        cputs(value);
        gotoxy(1,row-Lrecd->rowOrigin+1);
        textcolor(Lcolab);
        if ((strlen(value) > 0) && (value[0]!=' ')) putch(value[0]);
        else putch(BLOTCH);                     /* to show up a space char */
        }
      else {
        N = 1 + strlen(value)/Lcolm->columnWidth;
        if (N>1) {       /* display could have been on multiple lines */
          value[Lcolm->columnWidth] = '\0';
          cputs(value);
          for (line = row+1; line <= row+N-1, line < Lrecd->displayed; line++) {
            ListGetString(Lcolm,line,value);
            value[Lcolm->columnWidth] = '\0';
            gotoxy(1,line-Lrecd->rowOrigin+1); clreol();
            cputs(value);
         }}
        else cputs(value);
      }
      gotoxy(1,row-Lrecd->rowOrigin+1);  /* move to start of line */
    }
} /* end Listline */

void ListaColumn(int column)
     /* display all records of <column> if column visible */
{   int line;
    ListWindow(column);
    if (Lcolm->visible) {
      for (line = Lrecd->rowOrigin; line < Lrecd->rowOrigin+Lrecd->displayed; line++)
          Listline(line,Lcolt,false);
    }
}  /* end ListaColumn */

void ListDisplayValues(void)
     /* display all fields of active list */
{   int C;
    for (C = 0; C <= Lrecd->topColumn; ListaColumn(C++));
}  /* end ListDisplayValues */


int ListLR(int L, int activeCol)
    /* determine the screen position of the columns of a list
       hide columns that lie off screen
       <L> is the left edge of the list in screen columns
       return the right edge of list */
{   int R,I,lastV,C;
    int tC;                    /* Lrecd->topColumn */
    int prevL,prevW;           /* previous column left and width */
    tC = Lrecd->topColumn;
       /* start at first column */
    Lcolm = Lrecd->List[0];
    Lcolm->visible = 1;
    lastV = 1;
    prevW = Lcolm->columnWidth;
    prevL = 1;
    Lcolm->columnLeft = prevL + L;
       /* find column lefts by adding column widths plus separators */
    R = 0;
    for (I = 1; I <= tC; I++) {
      Lcolm = Lrecd->List[I];
      prevL += prevW + 1;
      Lcolm->columnLeft = prevL + L;
      Lcolm->visible = 1;            /* make sure all cols visible */
      prevW = Lcolm->columnWidth;
         /* flag if active column has some part off screen */
      if ((I == activeCol) && (Lcolm->columnLeft + Lcolm->columnWidth > screenRight))
        R = Lcolm->columnLeft + Lcolm->columnWidth;
    }
       /* hide columns (except first) until active col is on screen */
    if (R > 0) {
      I = 1;
      Lcolm = Lrecd->List[I];
      C = Lcolm->columnLeft;
      do {
        Lcolm->visible = 0;
        I++;
        Lcolm = Lrecd->List[I];
      } while ((I < activeCol) && (R - screenRight >= Lcolm->columnLeft - C));
         /* reset column lefts */
      R = Lcolm->columnLeft - C;
      for (C = I; C <= tC; C++) {
        Lcolm = Lrecd->List[C];
        Lcolm->columnLeft -= R;
      }
    }
       /* hide columns off-screen to right */
    if (activeCol < tC) {
      I = activeCol + 1;
      do {
        Lcolm = Lrecd->List[I];
        I++;
      } while ((I <= tC) && (Lcolm->columnLeft + Lcolm->columnWidth < screenRight));
      if (Lcolm->columnLeft + Lcolm->columnWidth >= screenRight)
        for (C = I-1; C <= tC; C++) Lrecd->List[C]->visible = 0;
    }
         /* find list right from last visible column */
    for (I = activeCol; I <= tC; I++) {
      Lcolm = Lrecd->List[I];
      if (Lcolm->visible) {
        R = Lcolm->columnLeft + Lcolm->columnWidth;
        lastV = I;
      }
    }
        /* trim last column if necessary */
    if (R > screenRight) {
      Lrecd->List[lastV]->columnWidth -= screenRight - R;
      R = screenRight;
    }
       /* make sure list wide enough for title */
    C = minInteger(strlen(Lrecd->listTitle) + L + 3, screenRight);
    if (R < C) {
      Lrecd->List[lastV]->columnWidth += C - R;
      R = C;
    }
       /* set the internal grid lines */
    sety[0] = 2;
    sety[1] = 999;
    if (strlen(Lrecd->listTitle) == 0)
      Lrecd->offset = 3;
    else {
      Lrecd->offset = 5;
      sety[1] = 4;
      sety[2] = 999;
    }
       /* store list bounds */
    Lrecd->listLeft   = L;
    Lrecd->listRight  = R;
    return R;
} /* end ListLR */


void ListRefill(int list, int origin, int display)
     /* redraws <list> from row <origin> with page size <display> */
{   int I,J,K,C;
    int left;
    left = Lrecd->listLeft;
       /* determine the vertical grid and write the column headers */
    for (I = 0; I <= Lrecd->topColumn; I++) setx[I] = 999; /* set none */
    K = 0;
    for (I = 0; I <= Lrecd->topColumn; I++) {
      Lcolm = Lrecd->List[I];
      if (Lcolm->visible) {
        C = Lcolm->columnLeft - left;
        gotoxy(C,Lrecd->offset-2); cputs(Lcolm->columnTitle);
        for (J = K; J < I; J++, K++) setx[J] = C-1;
      }
    }
       /* redraw the internal grid */
    DrawGridSet(list, 'S');
       /* rewrite the list title */
    textcolor(Lcolt);
    if (Lrecd->offset == 5) PrintCharWindow('C',1,Lcolt,Lcolb,Lrecd->listTitle);
       /* reset list controls */
    Lrecd->displayed = display;
    Lrecd->rowOrigin = origin;
       /* show the field entries */
    ListDisplayValues();
}  /* end ListRefill */


void ListOn(int list, int colm, int left,int top, char frame, int Tcolour, int Bcolour)
     /* display <list> with screen position <left>,<top>
        using <frame> type {single,double}
        making <colm> the active column
        <Tcolour> & <Bcolour> are text and background colours */
{   int L,R,T,B;         /* left,right, top,bottom bounds of list */
    ListOff(list);       /* make sure that any previous display is removed */
      /* find and set bounds */
    L = left; T = top;
    R = ListLR(L,colm);     /* set column colm active */
    B = Lrecd->qtyInList + T + Lrecd->offset;
    if (B > screenBottom-2) B = screenBottom-2;
    Lrecd->listTop    = T;
    Lrecd->listBottom = B;
       /* store covered screen area */
    gettext(L,T,minInteger(R,screenRight),B,Lrecd->Larea);
       /* make area a boxed window */
    DrawWinBox(list,L,T,R,B, frame, 'c', Tcolour,Bcolour);
       /* fill in the window with the list */
    ListRefill(list, 0, B-T-Lrecd->offset);
    listIsOn[list] = true;
    Lrecd->columnmax = minInteger(Lrecd->columnmax,Lrecd->topColumn);
} /* end ListOn */


void ListRedisplay(int list, int row, int column)
     /* redisplay <list> making <column> and <row> active */
{   int L,R,T,B;           /* left,right, top,bottom bounds of list */
    int b;
    Lrecd = listSet[list];
      /* get bounds and uncover buried screen */
    L = Lrecd->listLeft;
    R = minInteger(screenRight,Lrecd->listRight);
    T = Lrecd->listTop;
    B = Lrecd->listBottom;
    puttext(L,T,R,B,Lrecd->Larea);
      /* find and set new bounds and get screen area to be covered */
    R = ListLR(L,column);
    b = Lrecd->qtyInList + T + Lrecd->offset;
    if (b > screenBottom-2) b = screenBottom-2;
    Lrecd->displayed += b-B; B = b;
    Lrecd->listBottom = B;
    gettext(L,T,minInteger(R,screenRight),B,Lrecd->Larea);
       /* make new area a boxed window */
    DrawWinBox(list,L,T,R,B, box[list].style, 'c', box[list].Tcol,box[list].Bcol);
       /* fill in the window with the list */
    ListRefill(list, Lrecd->rowOrigin, Lrecd->displayed);
       /* set column and row as active */
    ListWindow(column);
    Listline(row,Lcolat,true);
} /* end ListRedisplay */


int MoveDisplayrow(int list, int atrow, int row, int column, int Refresh)
    /* changes the active row of <list>
       from <atrow> of active column to <row> of <column>
       <Refresh> = 1 forces a rewrite of all fields */
{
    Lrecd = listSet[list];
    if (Refresh == 1)
      ListDisplayValues();
    else
      Listline(atrow,Lcolt,false);

    ListWindow(column);
    Listline(row,Lcolat,true);
    return row;
} /* end MoveDisplayrow */


int MoveAtListrow(int row, int column, int newrow, int Refresh)
    /* changes row of the active list from <row> to <newrow> of <column>
       <Refresh> true forces a rewrite of all fields */
{   int I;
       /* if newrow not already displayed
          OR refresh requested then rewrite list fields */
    if ((newrow < Lrecd->rowOrigin)
    || (newrow > Lrecd->rowOrigin+Lrecd->displayed-1) || (Refresh)) {
      if (abs(row-newrow) == Lrecd->displayed-1)
          I = Lrecd->rowOrigin-row+newrow;
      else
          I = newrow-((Lrecd->displayed-1)/2);
      Lrecd->rowOrigin = maxInteger(0,minInteger(I,newrow-Lrecd->displayed+1));

      ListDisplayValues();
      ListWindow(column);
      }
    else /* deactivate row */
      Listline(row,Lcolt,false);
       /* make newrow active */
    Listline(newrow,Lcolat,true);
    return newrow;
} /* end MoveAtListrow */


int MoveAtListcolumn(int list, int row, int column, int newcolumn)
    /* changes column of <list> from <column> to <newcolumn>
       making <row> of <newcolumn> active */
{
    if (newcolumn != column) {    /* needs doing */
      if (Lrecd->List[newcolumn]->visible == 0)  /* new column hidden */
        ListRedisplay(list, row, newcolumn);
      else {
        Listline(row,Lcolt,false); /* deactivate row */
        ListWindow(newcolumn);     /* change active column */
        Listline(row,Lcolat,true); /* activate row */
      }
    }
    return newcolumn;
} /* end MoveAtListcolumn */


void ListRestart(int list, int column)
     /* make <list> active on <column> */
{   Lrecd = listSet[list];
    Lcolt = box[list].Tcol;
    Lcolb = box[list].Bcol;
    Lrecd->displayed = Lrecd->listBottom-Lrecd->listTop-Lrecd->offset;
    SetTextColour(Lcolt,Lcolb);
    ListWindow(column);
} /* end ListRestart */


void StoreCut(int row)
     /* makes a copy of a list slot */
{
    RowDispose(Lcolm,cutItem);                 /* discard prior cut */
    cutOrigin = ListOrigin[row];               /* the original row */
    ListitemDuplicate(Lcolm, row, cutItem);
}  /* end StoreCut */

void PasteCut(int row)
     /* places cut item in a list slot */
{
    if (Lcolm->item[cutItem] != NULL) {
      RowDispose(Lcolm, row);
      Lcolm->item[row] = Lcolm->item[cutItem];
      ListOrigin[row] = cutOrigin;               /* the original row */
      Lcolm->item[cutItem] = NULL;               /* empty store */
    }
}  /* end PasteCut */

void DelCut(int list, int column, int row)
     /* deletes a row from a single column */
{   int I;
    StoreCut(row);                         /* keep cut item */
    ListDefaultFill(list, column, column, Lcolm->qtyInColumn,0); /* add empty row */
    for (I = row; I < Lcolm->qtyInColumn; I++) { /* scroll down pointers */
        Lcolm->item[I] = Lcolm->item[I+1];
        ListOrigin[I] = ListOrigin[I+1];   /* keep track of original source */
    }
    Lcolm->qtyInColumn--;                  /* remove added row */
}  /* end DelCut */

void InsCut(int list, int column, int row)
     /* insert a default row into a single column */
{   int I;
    Lcolm->qtyInColumn += 1;
    if (Lcolm->qtyInColumn > Lrecd->qtyInList) Lrecd->qtyInList++;
    for (I = Lcolm->qtyInColumn-1; I > row; I--) {
      Lcolm->item[I+1] = Lcolm->item[I];
      ListOrigin[I+1] = ListOrigin[I];  /* keep track of original source */
    }
    ListitemDuplicate(Lcolm, row, row+1);
    ListDefaultFill(list, column, column, row,0);
}  /* end InsCut */


void ShowChar(int where, char thechar, int colour, int keep)
     /* write one character <thechar> in active field
        at field column <where> in <colour> */
{   int w,x,y;
    w = Lcolm->columnWidth;
    x = 1 + (where-1) % w;
    y = 1 + (where-1) / w;
    gotoxy(x,y);
    textcolor(colour);
    if ((thechar != ' ') || (colour == Lcoledit))
      putch(thechar);
    else                 /* to show up a space char */
      putch(BLOTCH);

    if (keep) gotoxy(x,y);
}  /* end ShowChar */


void renew(char *Ttemp, int startchar, int limit)
     /* rewrites a field with text <Ttemp>
        from the <startchar> char of <Ttemp>
        for no more than <limit> chars */
{   char string[101];
    int i;
    clrscr();            /* erases current field display */
    textcolor(Lcoledit);
    limit = minInteger(limit,strlen(Ttemp)-startchar);
    for (i = 0; i < limit; i++) string[i] = Ttemp[i+startchar];
    string[limit] = '\0';
    cputs(string);
} /* end renew */


void straddchr(char *Ttemp, char editkey, int used)
     /* adds char <editkey> to end of <Ttemp> */
{   Ttemp[used]   = editkey;
    Ttemp[used+1] = 0;
}  /* end straddchr */


void ListEditField(unsigned char Ttype, double Tmin, double Tmax, int verify,
               unsigned char editkey, char *usertext)
     /* edits the displayed field entry of active list column row
        <Ttype>    is the form
        <Tmin>,<Tmax> are the min & max limits for the column data
        <verify>   is the verification code
        <editkey>  is the key that the user pressed to initiate the edit
                   input >127 to keep initial usertext
        <usertext> is the initial and final field entry as text */
{
    int   err;            /* error code */
    int   P;              /* local index */
    char  Ttemp[255];     /* text being edited */
    int   I;              /* field display position of active char */
    int   used;           /* length of Ttemp */
    int   limit;          /* maximum number of visible chars */
    int   startchar = 0;  /* first char of Ttemp visible (for scrolling) */
    char  inlo,inhi;      /* lowest and highest valid ASCII char for data form */
    unsigned char inchar; /* the latest keyboard char read */
    double Rtemp;         /* dummy value for testing numeric entries */
    int   needRead;       /* true if keyboard needs to be read */
    int   Nullkey;        /* true if a two-byte key pressed */

       /* set limits */
    if ((wrap) && (Ttype > 8))
      limit = Lcolm->columnWidth*(2 + ((int)Tmax - 1)/Lcolm->columnWidth);
    else if (wrap)
      limit = 2*Lcolm->columnWidth-1;
    else
      limit = Lcolm->columnWidth-1;
    switch (Ttype) {
      case  2 : ;
      case  4 : inlo = '-'; inhi = '9'; break;
      default : inlo = ' '; inhi = '~'; break;
    }
       /* establish text to edit */
    if (editkey <= 127) {           /* new text wipes out current text */
      if (editkey >= ' ') {         /* printable char OK */
        straddchr(Ttemp,editkey,0);
        I = 2;
        }
      else {                        /* control char ignored */
        Ttemp[0] = 0;
        I = 1;
      }
      needRead = true;
      }
    else {                          /* use current field value */
      strcpy(Ttemp,usertext);
      needRead = false;             /* editkey is a special key */
      inchar = editkey;
      I = 1;
    }
    used = strlen(Ttemp);
    if (Ttype >= TEXTTYPE)
      PrintCharScreen('L',2, GREEN,BLUE," at len max ");

       /* error return loop */
    do {
      renew(Ttemp,startchar,limit);
      editkey = 255; outkey = 0;     /* set empty return */
      Nullkey = false;               /* a single keycode */

         /* main editing loop */
      do {
        if (Ttype >= TEXTTYPE)
          Print3IntScreen(1,3, GREEN,BLUE, I+startchar,used,Tmax-1, 3,3,3);
           /* if active char not at end, show it */
        ShowChar(I,Ttemp[I+startchar-1], Lcolat,1);
           /* wait for char from user */
        if (needRead) inchar = getch();
           /* if active char visible, move to it */
        /* if (I+startchar < used) */
          ShowChar(I,Ttemp[I+startchar-1], Lcolat,0);
           /* process user key */
        if ((inchar == BACKSPACE) && (I > 1)) {  /* delete previous char */
            I--;
            used--;
            for (P = I+startchar-1; P < used; P++) Ttemp[P] = Ttemp[P+1];
            Ttemp[used] = 0;
            renew(Ttemp,startchar,limit);        /* redisplay shifted text */
          }
        else if ((inchar>=inlo) && (inchar <= inhi)) {  /* valid char */
          if ((textInsert==0) && (I <= used-startchar)) {  /* insert mode on */
            for (P = ++used; P > I+startchar-1; P--) Ttemp[P] = Ttemp[P-1];
            Ttemp[I+startchar-1] = inchar;
            renew(Ttemp,startchar,limit);        /* redisplay shifted text */
            }
          else {                                 /* overwrite mode on */
            if (I+startchar <= used)             /* overwrite */
              Ttemp[I+startchar-1] = inchar;
            else                                 /* add to end */
              straddchr(Ttemp,inchar,used++);
            if (I < limit)
              ShowChar(I,Ttemp[I+startchar-1], Lcoledit,0); /* show altered char */
          }
          if (I < limit) I++;                    /* still visible */
          else renew(Ttemp,++startchar,limit);   /* keep active text in view */
          }
        else if ((inchar == 0) || (inchar > 128)) {  /* double keycode */
          if (needRead)
              inchar = getch();                  /* get second keycode */
          else
              inchar = inchar-128;               /* decode second keycode */

          switch (inchar) {
                 /* move to first char */
            case HOME : if (I+startchar > 1) {
                          I = 1; startchar = 0;
                        } break;
                 /* move left one char */
            case LARROW : if (I > 1) I--;
                        else if (startchar > 0) startchar--; break;
                 /* move right one char */
            case RARROW : if (I+startchar <= used) {
                          if (I < limit) I++;
                          else startchar++;
                        } break;
                 /* move to last char */
            case END  : if (I+startchar <= used) {
                          if (used-startchar <= limit)
                             I = used-startchar;
                          else {
                             I = limit; startchar = used-limit;
                          }
                        } break;
                 /* toggle insert/overwrite mode */
            case INS  : textInsert = 1-textInsert; break;
                 /* delete active char */
            case DEL  : if ((used > 0) && (I+startchar <= used)) {
                          used--;
                          for (P = I+startchar-1; P < used; P++) Ttemp[P] = Ttemp[P+1];
                          Ttemp[used] = 0;
                        } break;
                 /* control key belongs to calling routine */
            default   : editkey = 0; outkey = 256*inchar;
                        Nullkey = true;
          }
          renew(Ttemp,startchar,limit); /* visible text might need changing */
        }
        needRead = true;
        }
        /* watch for a "leave this field" key */
      while ((inchar != ENTER) && (inchar != ESC)
         &&  (inchar != TAB) && (inchar != SHIFTTAB)
         &&  (!Nullkey));

         /* check that entered text conforms to data form */
      err = 0;
      if (inchar == ESC) {            /* abandon changes */
        strcpy(Ttemp,usertext);       /* make edit text = initial text */
        renew(Ttemp,startchar,limit);
        }
      else {                          /* test new field value */
           /* if not empty, clean up text */
        if (used > 0) {
          while ((used > 1) && (Ttemp[used-1] == ' ')) used--;
          P = 1; while ((P < used) && (Ttemp[P-1] == ' ')) P++;
          if (P > 1) {
            strcpy(Ttemp,Ttemp+P);
            used = used-P+1;
          }
          Ttemp[used] = 0;
        }
           /* test for valid data form */
        if (Ttype >= TEXTTYPE) {
          if (used < Tmin) err = 1;
          else if (used > Tmax-1) err = 2;
          }
        else {
          if ((used == 0) || (Ttemp[0] == ' ')) {
            Rtemp = NOVALUE;    /* if empty, default is 'not specified' */
            Ttemp[0] = '\0';
            }
          else {
            Rtemp = atof(Ttemp);
            if (Rtemp < Tmin) err = 4;
            else if (Rtemp > Tmax) err = 5;
          }
        }
           /* apply verification adjustments */
        if (err == 0) err = Verify(verify,Ttemp);
      }
         /* print/remove error message */
      if (err != 0)
        PrintCharScreen('C',screenBottom-1,RED,WHITE,errMsg[err]);
      else
        PrintCharScreen('C',screenBottom-1,BLACK,BLACK,errMsg[0]);
      }
    while (err != 0);

    if (Ttype >= TEXTTYPE) {
      PrintCharScreen('L',2, BLACK,BLACK, "              ");
      PrintCharScreen('L',3, BLACK,BLACK, "              ");
    }
    strcpy(usertext,Ttemp);
    if (editkey == 255) outkey = inchar;
} /* end ListEditField */


unsigned int DoEdit(int row, int column, unsigned char option, int list)
     /* sets up a field for editing and restores list edit on return
        field is in <list> at <column> <row>
        <option> is the keycode to sent into the edit
        returns key used to leave field */
{   char tempstr[LISTLEN];  /* the field value as text */
    int Y1,Y2;              /* the edit rows as screen rows */
    int line;               /* row index */
    ListGetString(Lcolm,row,tempstr);
       /* make the field into the active window */
    Y1 = Lrecd->listTop+Lrecd->offset+row-Lrecd->rowOrigin;
    if (wrap) {
      if (Lcolm->Cform > 8) { /* text */
        Y2 = Y1 + ((int)Lcolm->Cmax - 1)/Lcolm->columnWidth;
        if (column > 1) Y2++; }
      else                   /* numeric */
        Y2 = Y1 + 1;
      if (Y2 > Lrecd->listBottom-1) Y2 = Lrecd->listBottom-1;
      }
    else Y2 = Y1;
    window(Lcolm->columnLeft,Y1,Lcolm->columnLeft+Lcolm->columnWidth-1,Y2);
       /* send field value (tempstr) to editor */
    textbackground(BLACK);
    ListEditField(Lcolm->Cform,Lcolm->Cmin,Lcolm->Cmax,Lcolm->Cverify,option,tempstr);
    ListRestart(list,column);
       /* store edited value */
    if (outkey != ESC) {
      saved = false;
      if (Lcolm->Cform > 8) FillListRow(Lcolm,row,tempstr, 0.0);
      else if (tempstr[0] != '\0') FillListRow(Lcolm,row,"", atof(tempstr));
      else FillListRow(Lcolm,row,"", NOVALUE);
    }
       /* restore column display (removes edit overwrites down column) */
    for (line = row+1; line <= row+Y2-Y1; line++)
        Listline(line,Lcolt,false);
        Listline(row,Lcolt,false);   /* leave cursor on <row> */
    return outkey;
} /* end DoEdit */


int ListEdit(int list, int left, int top, int column, int row, int style)
    /* general list editor
       edit without regard to the source of the data
       display then edit <list> at position <left> <top>
       making the initial active <column> and <row>
       <style> 0 : ENTER means pick active for return
               1 : ENTER means move down a row
               2 : use list as a menu, ie NO EDITING ALLOWED
               3 as for 2 but return after row change
               4 : insert/move/swap of data allowed (NB row = last row) */
{   int  chosen;
    unsigned char option;     /* user input key */
    int  atcolumn,atrow, I,C;
    int  Refresh = true;
    unsigned int userkey = 0;
    int  changes = false;
    char message[5][60] =      /* user help */
         { "Enter selects entry;  Esc picks none",
           "Enter moves down a row;  Esc cancels changes or exits",
           "Enter selects entry;  Esc exits",
           "Scroll to see selection;  Esc selects entry",
           "Re-arrange values: Del,Ins,Cut,Paste ;  Esc to accept" };

    if (listSet[list] != NULL) {
      Lrecd = listSet[list];
      chosen = false;
      if (column < Lrecd->columnmin) column = Lrecd->columnmin;
      if (column > Lrecd->columnmax) column = Lrecd->columnmax;
         /* fill init row numbers */
      if (style == 4) {
        Lcolm->item[cutItem] = NULL;      /* no cut item on entry */
        for (I = 0; I < row; ListOrigin[I] = I++);
        for (I = row; I < LISTLIMIT; ListOrigin[I++] = -1);
        row = 0;
      }
         /* make sure up-to-date list is visible */
      if (listIsOn[list]) ListRestart(list,column);   /* list already visible assumes no changes */
      else ListOn(list,column, left,top, 'D', Lcolt, Lcolb); /* display list */
         /* user help line */
      PrintCharScreen('C',screenBottom,WHITE,BLACK,message[style]);
         /* make sure active row/column is current */
      atrow = row;
      atcolumn = column;
      ListWindow(atcolumn);
      MoveAtListrow(atrow,atcolumn,row,Refresh); Refresh = false;
      MoveAtListcolumn(list,atrow,atcolumn,column);
         /* process user action */
      do {
        if (userkey == 0) option = getch();
        else { option = userkey % 256; userkey /= 256; }
        if (option == 0) {
          if (userkey == 0) option = getch();  /* get double key */
          else { option = userkey; userkey = 0; }
          switch (option) {
            case UPARROW: if (row > 0) row--; else row = Lcolm->qtyInColumn-1;
                          break;
            case DNARROW: ;
            case ENTER  : if (row < Lcolm->qtyInColumn-1) row++; else row = 0;
                          break;
            case RARROW : if (style < 2) {
                          userkey = DoEdit(row,column,option+128,list);
                          if (outkey != ESC) changes = true;
                          } break;
            case LARROW : ;
            case SHIFTTAB: if (column > Lrecd->columnmin) column--; else column = Lrecd->columnmax;
                          break;
            case HOME   : row = 0; break;
            case END    : row = Lcolm->qtyInColumn-1; break;
            case PAGEUP : if (row > Lrecd->displayed-1) row = row-Lrecd->displayed-1;
                          else row = 0; break;
            case PAGEDN : if (row < Lrecd->qtyInList-Lrecd->displayed) row = row+Lrecd->displayed;
                          else row = Lcolm->qtyInColumn-1; break;
            case INS    : if (style < 2) {
                            Lrecd->qtyInList++;
                            for (C = 0; C <= Lrecd->topColumn; C++) {
                              Lcolm = Lrecd->List[C];
                              Lcolm->qtyInColumn = Lrecd->qtyInList;
                              for (I = Lrecd->qtyInList-1; I > row; I--) Lcolm->item[I+1] = Lcolm->item[I];
                              ListitemDuplicate(Lcolm, row, row+1);
                            }
                            Lcolm = Lrecd->List[column];
                            ListDefaultFill(list, 0,Lrecd->topColumn, row,1);
                          }
                          else if (style == 4)   /* location editing */
                            InsCut(list,column,row);
                          Refresh = true;
                          break;
            case DEL    : if (style < 2) {
                            if (Lrecd->qtyInList == 1) {
                              Lcolm = Lrecd->List[column];
                              ListDefaultFill(list, 0,Lrecd->topColumn, row,1);
                              }
                            else {
                              Lrecd->qtyInList--;
                              for (C = 0; C <= Lrecd->topColumn; C++) {
                                Lcolm = Lrecd->List[C];
                                Lcolm->qtyInColumn = Lrecd->qtyInList;
                                RowDispose(Lcolm,row);
                                for (I = row; I < Lrecd->qtyInList-1; I++) Lcolm->item[I] = Lcolm->item[I+1];
                              }
                              if (row >= Lrecd->qtyInList) row--;
                            }
                          }
                          else if (style == 4)    /* location editing */
                            DelCut(list,column,row);
                          Refresh = true;
                          break;
            case F8     : chosen = true; break;
            default     : if ((style > 1) && (((option > 15) && (option < 26))
                          ||  ((option > 29) && (option < 39))
                          ||  ((option > 43) && (option < 51)))) {
                            chosen = true; row = LISTLIMIT+option;  }
                          else
                            putchar(BELL);
          }
         }
        else {
          userkey = 0;
          switch (option) {
            case TAB   : if (column < Lrecd->columnmax) column++; else column = Lrecd->columnmin;
                         break;
            case ENTER : if ((style == 1) && (row < Lcolm->qtyInColumn-1)) row++;
                         else chosen = true; break;
            case ESC   : chosen = true;
                         if (changes) {
                           ListWindow(atcolumn);
                           chosen = alertYesNo("DISCARD your edits?","Yes/No");
                         } break;
            default    : if ((style < 2)  /* value or text editing */
                         && ((option >= ' ') || (option == BACKSPACE))) {
                             userkey = DoEdit(row,column,option,list);
                             if (userkey != ESC) changes = true;
                         }
                         else if (style == 4) {  /* location editing */
                           switch (toupper(option)) {
                /* cut */    case 'T'||'C' :
                               StoreCut(row);
                               ListDefaultFill(list, column,column, row,0);
                               break;
                /* paste */  case 'P' :
                               PasteCut(row);
                               break;
                /* ins */    case 'I' :
                               InsCut(list,column,row);
                               break;
                /* del */    case 'D' :
                               DelCut(list,column,row);
                               break;
                             default  : putchar(BELL);
                           }
                           Refresh = true;
                         }
          }
        }
           /* move active location */
        if (row < LISTLIMIT) {
          if ((style == 3) && (row != atrow)) chosen = true;
          atrow = MoveAtListrow(atrow,atcolumn,row,Refresh);
          Refresh = false;
          atcolumn = MoveAtListcolumn(list,atrow,atcolumn,column);
        }
      } while (!chosen);
      if (option == ESC) row = -1;
      RowDispose(Lcolm,cutItem);                      /* clear copy */
    }
    PrintCharScreen('l',screenBottom,WHITE,BLACK,""); /* clear line */
    return row;
} /* end ListEdit */


void SortLR(int l, int r)
     /* a QuickSort of list entries in a list column */
{   int   i,j;
    void  *y;
    if (l < r) {
      i = l; j = r;
      ListGetString(Lcolm, (i+j)/2, SortPivot);
      do {
          ListGetString(Lcolm,i,SortTest);
          while (strcmp(SortTest,SortPivot) < 0) {
              i++; ListGetString(Lcolm,i,SortTest); }
          ListGetString(Lcolm,j,SortTest);
          while (strcmp(SortTest,SortPivot) > 0) {
              j--; ListGetString(Lcolm,j,SortTest); }
          if (i <= j) {
            y = Lcolm->item[i];
            Lcolm->item[i] = Lcolm->item[j];
            Lcolm->item[j] = y;
            i++; j--;
          }
        }
      while (i < j);
      SortLR(l,j);
      SortLR(i,r);
    }
} /* end SortLR */

void SortList(int l, int r)
     /* padding needed by C for sorting */
{   SortTest = (char *) malloc(PATHSIZE);  /* allocate RAM */
    SortPivot = (char *) malloc(PATHSIZE);
    SortLR(l,r);                    /* do sort */
    free(SortTest);                 /* free RAM */
    free(SortPivot);
} /* end SortList */


int  GetFileList(int list, char *title, char *fextn)
     /* create a list of files found on the system
        into <list> with a heading of <title>
        for all files matching *.<fextn>
        a <fextn> of "*" means all data files */
{   int  i,posn,found;
    char filename[PATHSIZE], name[PATHSIZE];
    int  addExn;               /* true if display file-extension */
    struct ffblk DosFileRec;

    ListNew(list,title);
    ListFillInit(list,0,"file name", TEXTTYPE, 3+PLATFORM, 0,12,0);
    addExn = ((strpos(fextn,'*') >= 0) || (strpos(fextn,'?') >= 0));
    strcpy(filename,workpath);
    strcat(filename,"*");
    strcat(filename,fextn);
    found = findfirst(filename,&DosFileRec,0);
    while (!found) {
      strcpy(name,DosFileRec.ff_name);
      if (name[0] != '.') {                    /* not a special DOS file */
        posn = strpos(name,'.');
        if ((posn >=0) && (addExn)) {
          for (i = 0; i<5; i++) filename[i] = name[posn+i];
          if ((strcmp(filename,LIBEXTN))       /* extension rejects */
          &&  (strcmp(filename,DFNEXTN))
          &&  (strcmp(filename,GDFEXTN))
          &&  (strcmp(filename,EXEEXTN))) {
            strcpy(filename,name);
            ListFill(list,0,Lcolm->qtyInColumn, filename, 0);
          }}
        else if (strstr(name,fextn) != NULL)   /* requested extension */
        {
          if (posn == -1) posn = strlen(name);
          strncpy(filename,name,posn);
          filename[posn] = '\0';
          ListFill(list,0,Lcolm->qtyInColumn, filename, 0);
        }
      }
      found = findnext(&DosFileRec);
    }
    SortList(0,Lrecd->qtyInList-1);
    return Lrecd->qtyInList;
} /* end GetFileList */


void FileListChoice(int list, int start, char *title, char *fileExn)
     /* get and display a list of files, allowing the user to pick one
        using <list> with a heading of <title>
        for all files matching *.<fileExn>
        make the initial "active" list entry = <start>
        return selection in <title> */
{   int qty,row,I;
    char fname[PATHSIZE];
    struct text_info winInfo;
       /* save screen image */
    gettextinfo(&winInfo);
       /* get file list */
    qty = GetFileList(list,title,fileExn);
       /* if files exist then display for choice */
    if (qty == 0)
      strcpy(title,"?");
    else {
         /* note option to type in an entry */
      alert('L',"Press Esc for none","Alt+U to type in a filename");
         /* ask user for name */
      Lcolt = BLACK;
      Lcolb = WHITE;
      if (start > qty) row = qty; else row = start;
      if (qty < 20) I = 3; else I = 2;
      row = ListEdit(list,6,I,0,row,2);
         /* get user name */
      if (row == -1)                 /* Esc-key pressed */
        title[0] = '\0';
      else if (row >= LISTLIMIT) {   /* Alt-key used */
        alertClear();
        strcpy(title,alert('S',"Enter the required filename",""));
        }
      else {                         /* an entry selected */
        ListGetString(Lcolm, row, fname);
        sprintf(title,"%s%s",workpath,fname);
      }
         /* restore screen */
      ListOff(list);
      window(winInfo.winleft,winInfo.wintop,winInfo.winright,winInfo.winbottom);
    }
    alertClear();
} /* end FileListChoice */

char *FileList(char *title, char *extn, int start)
     /* let the user select from a list of files ending in <extn>
        the cursor is started on row <start>.
        uses list = 3 */
{   if (title[0] == '\0')
      strcpy(filename,"FILE LIST");
    else
      strcpy(filename,title);

    FileListChoice(3,start, filename, extn);
    return filename;
}  /* end FileList */


int ListGetQty(int list)
     /* returns the rows allocated to a list */
{
    return listSet[list]->qtyInList;
}

void TrimZero(char *tempstr)
     /* removes trailing zeros from a decimal string */
{   int P,Z;
    P = strpos(tempstr,'.');
    Z = strlen(tempstr);
    while ((Z > P) && (tempstr[Z-1] == '0')) Z--;
    if (P == Z-1) Z = P;
    tempstr[Z] = '\0';
}

void ListGetString(struct ListColumn *Lcolm, int row, char *string)
     /* returns the string value of a list field */
{
    switch (Lcolm->Cform) {
      case  2 : rD2 = (struct integer2 *) Lcolm->item[row];
                sprintf(string,"%d",rD2->content); break;
      case  4 : rD4 = (struct float4 *) Lcolm->item[row];
                if (rD4->content == NOVALUE) string[0] = '\0';
                else {
                  sprintf(string,"%.*f",Lcolm->Cdeci,rD4->content);
                  if (Lcolm->Cdeci > 0) TrimZero(string);
                  }
                break;
      case  8 : rD8 = (struct double8 *) Lcolm->item[row];
                if (rD8->content == NOVALUE) string[0] = '\0';
                else {
                  sprintf(string,"%.*f",Lcolm->Cdeci,rD8->content);
                  if (Lcolm->Cdeci > 0) TrimZero(string);
                  }
                break;
      case  9 : rD9 = (struct string9 *) Lcolm->item[row];
                strncpy(string,rD9->content,Lcolm->Cform); break;
      case 21 : rD21 = (struct string21 *) Lcolm->item[row];
                strncpy(string,rD21->content,Lcolm->Cform); break;
      case 41 : rD41 = (struct string41 *) Lcolm->item[row];
                strncpy(string,rD41->content,Lcolm->Cform); break;
      case 101: rD101 = (struct string101 *) Lcolm->item[row];
                strncpy(string,rD101->content,Lcolm->Cform);
    }
    if (Lcolm->Cform > 8) string[Lcolm->Cform-1] = '\0';
} /* end ListGetString */


double ListGetValue(struct ListColumn *Lcolm, int row)
     /* returns the numerical value of a list field */
{   double value;
    switch (Lcolm->Cform) {
      case  2 : rD2 = (struct integer2 *) Lcolm->item[row];
                value = rD2->content; break;
      case  4 : rD4 = (struct float4 *) Lcolm->item[row];
                value = rD4->content; break;
      case  8 : rD8 = (struct double8 *) Lcolm->item[row];
                value = rD8->content; break;
      default : value = 0.0;
    }
    return value;
} /* end ListGetValue */

int  ListGetOriginal(int row)
     /* returns the original row number before editing moved row values */
{
    return ListOrigin[row];
}

#endif