/* GDFmenu
      handles all the menu and submenu
      setup, display, selection, and preliminary processing
*/

#ifndef MENUMAX

#define MENUMAX     8   /* Number of main menu choices*/
#define SUBMENUMAX 10   /* Maximum number of submenu choices*/
#define MENUMAXLEN 20   /* Maximum length of a menu description */
#define SLABLELEN   5   /* shortcut label allowance */

short AtSubMenu;     /* active submenu of main menu */
enum boolean PDMoff; /* true if pull_down_menu is not visible */

     /* the menu colours */
const int COLOURTEXT    = BLACK;  /* Text colour */
const int COLOURBGROUND = WHITE;  /* Background colour */
const int COLOURACTIVE  = RED;    /* Current choice background colour */

char      Altset[52];                    /* Alt key converter */
short int SubMenuat[MENUMAX];            /* user last choice/auto choice */
char      MenuTitle[MENUMAX][SUBMENUMAX][MENUMAXLEN] ; /* menu text */
short int MenuForm[MENUMAX][SUBMENUMAX]; /* menu form eg dimmed */
int       MainPos[MENUMAX];              /* Main menu list start column */
int       SubMenulen[MENUMAX];           /* Maximum length of submenu names */
char      MenuCharAlt[MENUMAX];          /* Alt shortcut keys */
char      MenuChar[MENUMAX][SUBMENUMAX]; /* shortcut keys */
short int SubMenused[MENUMAX];           /* number of submenu items */
short int maintop;                       /* maximum main menu allocated */
char      option;                        /* from keyboard entry */
int       SubMenuArea[(SUBMENUMAX+2)*(MENUMAXLEN+SLABLELEN)]; /* for screen image under submenu */
int       SMleft,SMtop,SMright,SMbottom; /* submenu bounds */
short int menuat;                        /* active main menu */


/* prototypes */


void SetMenu(short main, short sub, short form, char *subtext, char subkey);
     /* sets menu entries */
  void SubMenuline(int subline, int colour);
       /* display the submenu entry for <subline> with <colour> */
  void SMbox(int height, int width, int kink, int A, int B, int C, int D, int E);
       /* draw the outline of the submenu display */
  void SubWindow(int X);
       /* display a pull down submenu at column X
          the main menu is given by the global menuat */
  void MoveSubMenuat(int MM, int SM);
       /* change the active submenu to SM for main menu MM*/
  void MainMenuitem(int menuat, int colour);
       /* display main menu menuat in colour */
  void MenuChosen(int I);
       /* change the selected main menu to I */
  char SubMenuKeyList(int AtM, char key);
       /* convert shortcut key to submenu AtM choice */
char MenuKey(char option, int AtM);
     /* converts a keyboard key into a submenu selection from main menu AtM */
int  MenuChoice(char option);
     /* actively select from menu */
void MenuWrite(void);
     /* displays the main menu line at top of screen */
void Restart(int list, char form);
     /* rewrites the screen after a call to HELP */
void Stubb(char *text);
     /* "programme incomplete" message */
unsigned int ProcessMenu(int list, char key);
     /* key activated menu selection */
void SetMenus(void);
     /* sets the initial menu items */

     /* the HELP displays */
void HelpASEG(void);
void HelpLib(void);
void HelpDefn(void);
void HelpRT(void);
void HelpAcro(void);
void HelpTrans(void);
void PutPair(int i, char *left, char *right);
     /* display edit function and key(s) to use */
void HelpEdit(void);


/* implimentation */


void SetMenu(short master, short sub, short form, char *subtext, char subkey)
     /* sets menu entries
        master = main menu to set
        sub    = submenu to set
        form   = display type {normal, dim, bright}
        subtext= submenu text to display
        subkey = shortcut key */
{ int I,P;
     /* check for valid request */
  master = abs(master);
  sub = abs(sub);
  if ((master < MENUMAX) && (sub < SUBMENUMAX)) {

    if ((sub == 0) || (master > maintop)) {  /* setting main menu OR new menu */
      if (master > maintop) {   /* new main menu */
        if (master == 0)                               /* first entry */
            MainPos[master] = 2;
        else for (I = maintop+1; I <= master; I++) { /* subsequent entry */
            MenuTitle[I][0][0] = '\0';               /* if left a gap */
            MainPos[I] = MainPos[I-1]+5+strlen(MenuTitle[I-1][0]);
        }
        for (I = maintop+1; I <= master; I++) {      /* set submenu as empty */
            SubMenused[I] = -1; SubMenuat[I] = -1;
        }
           /* keep track of highest allocated main menu */
        maintop = master;
           /* set menu location */
        MainPos[maintop+1] = screenInfo.screenwidth+10;
           /* set submenu shortcut keys as blank */
        for (I = 0; I < SUBMENUMAX; I++) MenuChar[master][I] = ' ';
        }
      else if (sub == 0) {      /* reset main menu */
           /* adjust menu locations */
        P = strlen(subtext)-strlen(MenuTitle[master][0]);
        for (I = master+1; I < maintop; I++) MainPos[I] += P;
           /* if necessary, set submenu as empty */
        if (strlen(subtext) == 0) {
          if (master == maintop)
          SubMenused[master] = -1; SubMenuat[master] = -1;
        }
      }
         /* set Alt_key for menu */
      MenuCharAlt[master] = subkey;
    }
       /* set (sub)menu title and form */
    strcpy(MenuTitle[master][sub],subtext);
    MenuForm[master][sub] = form;
       /* if a submenu item */
    if (sub > 0) {
      if (sub > SubMenused[master]) {  /* a new entry */
        if (sub > 1) for (I = SubMenused[master]+1; I < sub-1; I++) { /* fill any gap */
            MenuTitle[master][I][0] = '\0';
            MenuForm[master][sub] = 0;
        }
        SubMenused[master] = sub;   /* reset quantity used */
      }
         /* find maximum width of submenu entries */
      P = strlen(MenuTitle[master][1]);
      for (I = 2; I <= SubMenused[master]; I++)
        if (P < strlen(MenuTitle[master][I])) P = strlen(MenuTitle[master][I]);
      SubMenulen[master] = P;
         /* set submenu choice and shortcut key */
      SubMenuat[master] = -1;
      MenuChar[master][sub] = subkey;
    }
  }
}   /* end SetMenu */


  void SubMenuline(int subline, int colour)
       /* display the submenu entry for <subline> with <colour> */
  {   int I;
      int gap;  /* space between description and shortcut key */
      if (subline > 0) { /* not the main menu */
        textcolor(colour);
        gotoxy(3,subline+1);  /* leave left gap and allow for main menu */
        cputs(MenuTitle[menuat][subline]);
        gap = SubMenulen[menuat] - strlen(MenuTitle[menuat][subline]);
        for (I = 0; I <= gap; I++) putch(' ');
        putch(MenuChar[menuat][subline]);  /* shortcut key */
        textcolor(COLOURTEXT);             /* restore colour */
      };
    }  /* end SubMenuline */


  void SMbox(int height, int width, int kink, int A, int B, int C, int D, int E)
       /* draw the outline of the submenu display
          height,width = size
          kink is the column where the outline turns up to the main menu
          A    is the column for the unkinked vertical
          C,D  are the columns specifying the short horizontal
          B,E  are outline corner characters */
  {   int I,J;
      for (I = 1; I < height; I++) { gotoxy(A,I); putch(SVERTICAL); }
      for (I = 2; I < width; I++) { gotoxy(I,height); putch(SHORIZONTAL); }
      gotoxy(1,height); putch(SLOWLEFT);
      gotoxy(width,height); putch(SLOWRIGHT);
      J = width+1-A;
      for (I = 2; I < height; I++) { gotoxy(J,I); putch(SVERTICAL); }
      gotoxy(J,1); putch(B);
      for (I = C+1; I < D; I++) { gotoxy(I,1); putch(SHORIZONTAL); }
      gotoxy(kink,1); putch(E);
  } /* end SMbox */


  void SubWindow(int X)
       /* display a pull down submenu at column X
          the main menu is given by the global menuat */
  {
      int  I,lenT,moveleft,width,height,kink;
      if (PDMoff)  /* no curreny submenu is displayed */
        PDMoff = false;
      else         /* replace screen over Pull-Down-Menu area */
        puttext(SMleft,SMtop,SMright,SMbottom,SubMenuArea);

      if (SubMenused[menuat] == -1)  /* no entries in submenu */
        PDMoff = true;
      else {
           /* determin the submenu display area */
        SMtop    = 2;
        SMbottom = 3 + SubMenused[menuat];
        I    = SubMenulen[menuat] + SLABLELEN;
        lenT = strlen(MenuTitle[menuat][0]);
        if (X+I > screenInfo.screenwidth) {  /* offset to left */
          SMright = X + lenT - 1;
          SMleft  = SMright - I;
          kink    = X-SMleft+1;
          moveleft= 1;
          }
        else {                               /* offset to right */
          SMleft  = X;
          SMright = X + I;
          kink    = lenT;
          moveleft= 0;
        }
           /* store to be covered area then clear and set up as window */
        gettext(SMleft,SMtop,SMright,SMbottom,SubMenuArea);
        window(SMleft,SMtop,SMright,SMbottom);
        clrscr();
        window(SMleft,SMtop,SMright,SMbottom+1);
           /* draw outline */
        width  = SMright-SMleft+1;
        height = SMbottom-SMtop+1;
        if (moveleft)
          SMbox(height,width,kink,width,STOPLEFT,1,kink,SLOWRIGHT);
        else
          SMbox(height,width,kink,1,STOPRIGHT,kink,width,SLOWLEFT);
           /* write submenu */
        for (I = 1; I <= SubMenused[menuat]; I++) SubMenuline(I,COLOURTEXT);
           /* highlight the active submenu entry */
        if (SubMenuat[menuat] < 0) SubMenuat[menuat] = 1;
        SubMenuline(SubMenuat[menuat],COLOURACTIVE);
      };
    } /* end SubWindow */


  void MoveSubMenuat(int MM, int SM)
       /* change the active submenu to SM for main menu MM*/
  {
      if (SubMenuat[MM] != SM) {  /* only if changed */
           /* remove current highlight */
        if ((SubMenuat[MM] <= SubMenused[MM]) && (SubMenuat[MM] >= 0))
            SubMenuline(SubMenuat[MM],COLOURTEXT);
           /* add highlight */
        if ((SM <= SubMenused[MM]) && (SM > 0)) SubMenuline(SM,COLOURACTIVE);
           /* store which submenu is active for this main menu */
        SubMenuat[MM] = SM;
      }
  }  /* end MoveSubMenuat */


  void MainMenuitem(int menuat, int colour)
       /* display main menu menuat in colour */
  {      /* move to menu line */
      window(1,1,screenInfo.screenwidth,1);
         /* write text */
      textcolor(colour);
      gotoxy(MainPos[menuat],1);
      cputs(MenuTitle[menuat][0]);
         /* highlight Alt_key */
      gotoxy(MainPos[menuat],1);
      textcolor(COLOURACTIVE);
      putch(MenuTitle[menuat][0][0]);
         /* restore screen */
      textcolor(COLOURTEXT);
      window(1,2,screenInfo.screenwidth,screenInfo.screenheight);
  } /* end MainMenuitem */


  void MenuChosen(int I)
       /* change the selected main menu to I */
  {
      if (menuat < MENUMAX)           /* remove previous active menu */
          MainMenuitem(menuat,COLOURTEXT);
      menuat = I;                     /* keep selection */
      MainMenuitem(I,COLOURACTIVE);   /* show new active menu */
      SubWindow(MainPos[I]);          /* show submenu */
  } /* end MenuChosen */


  char SubMenuKeyList(int AtM, char key)
       /* convert shortcut key to submenu AtM choice */
  {   int I;
      I = strpos(MenuChar[AtM],key);
      if ((I < 0) || (I > SubMenused[AtM]))  /* key not valid */
        return 0;
      else {                                 /* found */
        SubMenuat[AtM] = I;
        return 1;
      }
  }  /* end SubMenuKeyList */


char MenuKey(char option, int AtM)
     /* converts a keyboard key into a submenu selection
        option is the key
        AtM    is the requested main menu */
{   int P;           /* index into string */
    char chosen = 0;
    if ((option != 0) && (option > 32))  /* single keycode */
      chosen = SubMenuKeyList(AtM,option);
    else if (option == 0) {              /* double keycode */
        option = getch();
        if (option > 50)                 /* not an Alt_key */
          chosen = SubMenuKeyList(AtM,option);
        else {                           /* Alt_key pressed */
          chosen = ((P = strpos(MenuCharAlt,Altset[option])) != -1);
          if (chosen) {                  /* valid Alt_key for a main menu */
            menuat = P;                  /* set new main menu choice */
            option = toupper(getch());   /* wait for submenu selection */
            if (option == 0) option = getch();  /* double keycode */
            chosen = SubMenuKeyList(menuat,option);
            }
          else {                         /* default choice */
            menuat = 0;
            SubMenuat[0] = 0;
          };
        };
    }

    if (!chosen) putchar(BELL);
    return chosen;
}  /* end MenuKey */


int MenuChoice(char option)
     /* actively select from menu
        active means highlighted menu entry
        selected means picked menu entry */
{   char chosen;               /* true (not 0) if a submenu selected */
    short int MM,SM;           /* local active menus; keep global values while cursing */
    int Mt;                    /* tempory main menu selection */
       /* store screen parameters at entry */
    struct text_info calling;
    gettextinfo(&calling);
       /* set menu conditions */
    textbackground(WHITE);
    PDMoff = true;             /* no submenu showing */
    menuat = MENUMAX;          /* equivalent to no menu active */
    chosen = 0;

       /* select menu */
    do {                       /* repeat until a selection made */
         /* update local active menus */
      if (menuat < MENUMAX) {  /* moving around menus */
        MM = menuat;
        SM = SubMenuat[MM];
        option = toupper(getch());  /* wait for selection */
        }
      else {                   /* first time in */
        MM = option;
        SM = 0;
        option = 1;
      }
      switch (option) {
        case 1 : option = MM;  MM = 0; SM = 0;  /* startup */
        case 0 : if (menuat < MENUMAX) option = getch(); /* double key-code */
             switch (option) {
               case LARROW  : if (MM > 0) MM--; else MM = maintop; break;
               case RARROW  : if (MM < maintop) MM++; else MM = 0; break;
               case UPARROW : if (SM > 1)
                                do SM--; while (strlen(MenuTitle[MM][SM]) == 0);
                              else
                                SM = SubMenused[MM];
                              break;
               case DNARROW : if (SM < SubMenused[MM])
                                do SM++; while (strlen(MenuTitle[MM][SM]) == 0);
                              else if (SM > 1)
                                SM = 1;
                              break;
               case HOME    : ;
               case PAGEUP  : if (SM > 1) SM = 1; break;
               case END     : ;
               case PAGEDN  : SM = SubMenused[MM]; break;
               default      : Mt = strpos(MenuCharAlt,Altset[option]);
                              if (Mt == -1) putchar(BELL);
                              else MM = Mt;
             }
             if (menuat != MM) MenuChosen(MM);  /* change of main menu */
             else MoveSubMenuat(MM,SM);         /* ?change of submenu */
             break;
        case ENTER : ;
        case ESC   : chosen = 1; break;
        default    : chosen = MenuKey(option,MM); /* shortcut key cf control key */
      }
    } while ((chosen == 0) && (menuat < MENUMAX));
       /*replace screen over SubMenu area*/
    if (!PDMoff) puttext(SMleft,SMtop,SMright,SMbottom,SubMenuArea);
       /*remove previous active option */
    MainMenuitem(menuat,COLOURTEXT);
    textattr(calling.attribute);

    if ((option != ESC) && (menuat < MENUMAX))
      return menuat*SUBMENUMAX + SubMenuat[menuat];
    else
      return 0;
}  /* end MenuChoice */


void MenuWrite(void)
     /* displays the main menu line at top of screen */
{   int I;
       /* clear screen */
    ClearScreen();
       /* make top strip */
    window(1,1,screenInfo.screenwidth,1);
    textbackground(COLOURBGROUND);
    clrscr();
       /* write main menu titles with Alt_key letter highlighted */
    for (I = 0; I <= maintop; I++) {
      gotoxy(MainPos[I],1);
      textcolor(COLOURTEXT);
      cputs(MenuTitle[I][0]);
      gotoxy(MainPos[I],1);
      textcolor(COLOURACTIVE);
      putch(MenuTitle[I][0][0]);
    }
       /* identify menu */
    textcolor(COLOURTEXT);
    PrintCharWindow('R',1,COLOURTEXT,COLOURBGROUND,"ASEG-GDF");
       /* set work area below menu */
    window(1,2,screenInfo.screenwidth,screenInfo.screenheight);
} /* end MenuWrite */

void _setcursortype(int cur_type);
void Restart(int list, char form)
     /* rewrites the screen after a call to HELP
        'list' is which of DRA was active when menu activated
        form = 'W' if wait for user to type a key */
{   int ans,i,L;
       /* wait for user to discard help */
    if (form == 'W') {
      _setcursortype(_NOCURSOR);       /* hide cursor */
      ans = getch();
      if (ans == 0) ans = getch();     /* discard any second keycode */
      _setcursortype(_NORMALCURSOR);   /* show cursor */
    }
       /* replace main menu */
    MenuWrite();
       /* replace DRA lists where last redisplay = active list */
    for (i = 1; i < 4; i++) {
      L = ((list+i) % 3);
      ListRedisplay(L,Lt[L].row,Lt[L].column);
    }
} /* end Restart */


void Stubb(char *text)
     /* stubb while programme is incomplete */
{   strcat(text," chosen");
    alert('A',text,"SORRY - option not yet available!");
}  /* end Stubb */


unsigned int ProcessMenu(int list, char key)
     /* key activated menu selection */
{ int choice;
  unsigned int returnCode = 1;
  char opt;
  char form = ' ';              /* form = 'W' if wait before refresh */
  choice = MenuChoice(key);
  if (choice > 5*SUBMENUMAX) {  /* set up help screen */
    DrawWinBox(3,1,1,screenRight,screenBottom-1, 'D','c',BLACK,WHITE);
    window(3,2,screenRight-2,screenBottom-1);
  }
  switch (choice) {
/* Libraries */
      case  1 : opt = 'n';                     /* new */
      case  2 : DiscardAll();                  /* get&load */
      case  3 : if (choice != 1) opt = 'g';    /* open&add */
                GetLib(opt); break;
      case  4 : SaveLib(activeLibrary,""); break;
      case  5 : SaveLib(activeLibrary," Saving as a new LIBRARY"); break;
      case  7 : if (!saved) {
                  if (alertYesNo("Changes not saved","Do you wish to save ?"))
                    SaveLib(-1," Saving as a new LIBRARY");
                }
                returnCode = 0; break;         /* quit */
/* Defn */
      case SUBMENUMAX+1 : GetDefn(""); DefnRelist(-1,0); break;
      case SUBMENUMAX+2 : MakeDefn(active[DEFNLIST]); break;
      case SUBMENUMAX+3 : DeleteDefn(); break;
/* Editing */
/*     case 2*SUBMENUMAX+1 : Stubb("Edit Undo"); break; */
      case 2*SUBMENUMAX+1 : returnCode = 256*DEL;        /* cut = copy+delete */
      case 2*SUBMENUMAX+2 : CopyDRA(list); break;        /* copy */
      case 2*SUBMENUMAX+3 : PasteDRA(list); break;       /* paste */
      case 2*SUBMENUMAX+4 : returnCode = 256*INS; break; /* insert */
      case 2*SUBMENUMAX+5 : returnCode = 256*DEL; break; /* delete */
/* Options */
           /* scroll horizontal window or wrap onto multiple lines */
      case 3*SUBMENUMAX+1 : wrap = false; break;
      case 3*SUBMENUMAX+2 : wrap = true;  break;
           /* show lists as all_defn/all_rt/all_acro */
/*      case 3*SUBMENUMAX+3 : Stubb("Options All fields"); break; */
           /* show lists as defn/rt/all_acro */
/*      case 3*SUBMENUMAX+4 : Stubb("Options Key fields only"); break; */
      case 3*SUBMENUMAX+3 : SetWorkPath(); break;
/* Transfer */
      case 4*SUBMENUMAX+1 : DisplayData(); break;
      case 4*SUBMENUMAX+2 : VerifyData(); break;
      case 4*SUBMENUMAX+3 : CreateFile(0); break;
      case 4*SUBMENUMAX+4 : CreateFile(1); break;
/* Help */
      case 5*SUBMENUMAX+1 : HelpASEG(); break;
      case 5*SUBMENUMAX+2 : HelpLib(); break;
      case 5*SUBMENUMAX+3 : HelpDefn(); break;
      case 5*SUBMENUMAX+4 : HelpRT(); break;
      case 5*SUBMENUMAX+5 : HelpAcro(); break;
      case 5*SUBMENUMAX+6 : HelpTrans(); break;
      case 5*SUBMENUMAX+7 : HelpEdit(); break;
  }
  if (choice > 5*SUBMENUMAX) form = 'W';
  if (choice > 4*SUBMENUMAX) Restart(list,form);
  return returnCode;
}  /* end ProcessMenu */


void SetMenus(void)
     /* sets the initial menu items */
{    /* the Alt_key (second keycode) converter */
  strcpy(Altset,"                QWERTYUIOP    ASDFGHJKL     ZXCVBNM");
     /* set the initial menu entries */
  maintop = -1;  /* updated by SetMenu */
  SetMenu(0,0,0,"Library",'L');
    SetMenu(0,1,0,"New",'N');
    SetMenu(0,2,0,"Load...",'L');
    SetMenu(0,3,0,"Open...",'O');
    SetMenu(0,4,0,"Save",'S');
    SetMenu(0,5,0,"save As...",'A');
    SetMenu(0,6,1,"",' ');
    SetMenu(0,7,0,"Quit",'Q');
  SetMenu(1,0,0,"Defn",'D');
    SetMenu(1,1,0,"Get",'G');
    SetMenu(1,2,0,"Make",'M');
    SetMenu(1,3,0,"Delete",'D');
  SetMenu(2,0,0,"Edit",'E');
    SetMenu(2,1,0,"Cut row",'T');
    SetMenu(2,2,0,"Copy row",'C');
    SetMenu(2,3,0,"Paste onto row",'P');
    SetMenu(2,4,0,"Insert a row",'I');
    SetMenu(2,5,0,"Delete row",'D');
  SetMenu(3,0,0,"Options",'O');
    SetMenu(3,1,0,"Scroll edit field",'S');
    SetMenu(3,2,0,"Wrap edit field",'W');
/*    SetMenu(3,3,0,"All fields",'A');
    SetMenu(3,4,0,"Key fields only",'K'); */
    SetMenu(3,3,0,"working Directory",'D');
  SetMenu(4,0,0,"Transfer",'T');
    SetMenu(4,1,0,"Display data",'D');
    SetMenu(4,2,0,"Verify  data",'V');
    SetMenu(4,3,0,"Export  GDF",'E');
    SetMenu(4,4,0,"Import  GDF",'I');
  SetMenu(5,0,0,"Help",'H');
    SetMenu(5,1,0,"ASEG-GDF",'G');
    SetMenu(5,2,0,"Libraries",'L');
    SetMenu(5,3,0,"Definitions",'D');
    SetMenu(5,4,0,"Record Types",'R');
    SetMenu(5,5,0,"Acronyms",'A');
    SetMenu(5,6,0,"Transfers",'T');
    SetMenu(5,7,0,"Editing",'E');
}  /* end SetMenus */


void HelpASEG(void)
{
    PrintCharWindow('C',1,BLACK,WHITE,"ASEG - GDF");
    gotoxy(66,1); cputs(GDFVERSION);
    PrintCharWindow('C',2,BLACK,WHITE,"The Australian Society of Exploration Geophysicists");
    PrintCharWindow('C',3,BLACK,WHITE,"General Data Format");
    gotoxy(1,5);
    cputs("This data format is designed to facilitate the exchange of data between\n\r");
    cputs("computing facilities.  At least two files are sent. The first is called the\n\r");
    cputs("DEFINITION (.DFN) file, which contains both the structure and information on");
    cputs("the content of the other files(s).\n\n\r");
    cputs("A DFN describes each possible data record structure (as a RECORD TYPE = RT)\n\r");
    cputs("The first 4 characters of each data record (or none for a null RT) are\n\r");
    cputs("those given under RT in the DFN.  For each RT, a set of ACRONYMs describe\n\r");
    cputs("each data field's FORMAT; NULL value; data UNITs; full NAME; and comments.");
}  /* end HelpASEG */

void HelpLib(void)
{
    PrintCharWindow('C',1,BLACK,WHITE,"LIBRARIES");
    gotoxy(1,3);
    cputs("You are allowed to have multiple sets of Definitions.  These are called\n\r");
    cputs("Libraries.  As you load Libraries into ASEG-GDF, they are given an index,\n\r");
    cputs("starting with 0 for the first, which is shown in the 'lib' column of the\n\r");
    cputs("Definition table.\n\n\r");
    cputs("Definitions can be added to a specific Library by changing the 'lib' value\n\r");
    cputs("as required before saving.  If you wish to amalgamate all loaded definitions");
    cputs("you do not have to change each 'lib' before saving.  Instead, use the MENU\n\r");
    cputs("option Library/SaveAs and answer 'Yes' to 'Should all definitions be saved\n\r");
    cputs("in this Library irrespective of original source library'.\n\n\r");
    cputs("If you use option Library/Save, ASEG-GDF will save all definitions back to\n\r");
    cputs("the Library with the 'lib' index as defined by the ACTIVE definition.");
}  /* end HelpLib */

void HelpDefn(void)
{
    PrintCharWindow('C',1,BLACK,WHITE,"DEFINITIONS");
    gotoxy(1,3);
    cputs("Each data file is composed of a set of records of one or more formats.\n\r");
    cputs("Each record format is defined in a record type within the definition.\n\r");
    cputs("Note that ONLY ONE definition describes ALL RECORDS of the file.\n\n\r");
    cputs("In data exchange, you send/receive a 'DFN' file as well as the data.\n\r");
    cputs("These are created/read by this program on data exchange.  If you wish,\n\r");
    cputs("you can make/get these 'DFN' files independantly of data exchange.\n\n\r");
    cputs("The 'description' field in the definition table is for clarification only.\n\r");
    cputs("It is stored in the Library but does not appear elsewhere.");
}  /* end HelpDefn */

void HelpRT(void)
{
    PrintCharWindow('C',1,BLACK,WHITE,"RECORD TYPES");
    gotoxy(1,3);
    cputs("A Record-Type is a label for a data record.  It is always 4 characters long.\n\r");
    cputs("This label is normally in upper case characters and is placed at the\n\r");
    cputs("beginning of each data record, left-justified and blank-filled to make up\n\r");
    cputs("the 4 characters required.\n\n\r");
    cputs("One Record-Type in a definition can be defined as 'missing' or 'null'.  This");
    cputs("is achieved by leaving the 'RT' entry blank.  The program will place '<>'\n\r");
    cputs("in the empty RT column to indicate the presence of a null.\n\n\r");
    cputs("You are not allowed to have duplicate RT labels in the same definition.\n\n\r");
    cputs("Each RT is described in detail by the ACRONYMS.  The Acronyms should not be\n\r");
    cputs("duplicated within an RT, but are often duplicated in different RT's.\n\n\r");
    cputs("The 'description' field in the RT table is for clarification only.  It is\n\r");
    cputs("stored in the Library but does not appear elsewhere.");
}  /* end HelpRT */

void HelpAcro(void)
{
    PrintCharWindow('C',1,BLACK,WHITE,"ACRONYMS");
    gotoxy(1,3);
    cputs("These describe the logical pieces of a record.  The ACRONYM label is from\n\r");
    cputs("1 to 8 characters long; it does not appear in the data record but is used\n\r");
    cputs("as a means of associating each logical piece of data with the control\n\r");
    cputs("parameters and to give the user of the data a reference key.\n\n\r");
    cputs("The FORMAT is written in simple FORTRAN code, only one format per entry.\n\r");
    cputs("This means that eg '2X,F12.3' is not allowed - it must be split into two\n\r");
    cputs("acronyms eg 'DUMMY with format A2' and 'INFO with format F12.3' or simply\n\r");
    cputs("'INFO with format F14.3' if the 2X are blanks.\n\n\r");
    cputs("The NULL entry defines the data value used when the data are missing.\n\r");
    cputs("The UNIT defines the data's unit of measurement.\n\r");
    cputs("The NAME describes the data; the ACRONYM should be an abbreviation of this.\n\r");
    cputs("COMMENTs are for any information about these data not placable above.\n\n\r");
    cputs("All table entries for acronyms are written to any definition (.DFN) file\n\r");
    cputs("created as well as being saved in the Library.");
}  /* end HelpAcro */

void HelpTrans(void)
{
    PrintCharWindow('C',1,BLACK,WHITE,"DATA TRANSFERS");
    gotoxy(4,3);
    cputs("DISPLAY DATA\n\r");
    cputs("Is used to look at a data file.  Usually this will be done in order to\n\r");
    cputs("construct a definition for a file of unknown format.  The active definition\n\r");
    cputs("can be used against the data to visually check both or either.\n\n\r");
    cputs("   VERIFY DATA\n\r");
    cputs("Is used to verify that data files obey the format of the active definition.\n\r");
    cputs("If a record fails the check, it is displayed with an indication of the\n\r");
    cputs("probable discrepency[s].\n\n\r");
    cputs("   EXPORT GDF\n\r");
    cputs("Is used to create ASEG-GDF (.GDF) files.  It uses the active definition.\n\r");
    cputs("Multiple source files can be merged but they should all fit into the export\n\r");
    cputs("definition record types.  A definition (.DFN) file is created.\n\n\r");
    cputs("   IMPORT GDF\n\r");
    cputs("Is used to create data files from ASEG-GDF (.GDF) files.  The data can be\n\r");
    cputs("selected in a different order and format to the original.  Not all data\n\r");
    cputs("need be selected either by acronyms or records.");
}  /* end HelpTrans */

void PutPair(int i, char *left, char *right)
     /* displays two messages on one line
        the left in black, the right in red */
{
    cputs(left);
    gotoxy(i,wherey());
    SetTextColour(RED,WHITE);
    cputs(right);
    SetTextColour(BLACK,WHITE);
    putch('\n');
    putch('\r');
}  /* end PutPair */

void HelpEdit(void)
{   int i = 19;
    PrintCharWindow('C',1,BLACK,WHITE,"EDITING");
    gotoxy(1,3);
    cputs("Movement keys\n\n\r");
    PutPair(i,"right one window","Ctrl+right_arrow");
    PutPair(i," left one window","Ctrl+left_arrow");
    PutPair(i,"right one column","tab");
    PutPair(i," left one column","shift_tab");
    PutPair(i,"   up one row",   "up_arrow");
    PutPair(i," down one row",   "down_arrow");
    PutPair(i,"   up one screen","Page_up");
    PutPair(i," down one screen","Page_down");
    PutPair(i,"   first row",    "Home");
    PutPair(i,"    last row",    "End");
    PutPair(i,"Insert a row",    "Ins");
    PutPair(i,"Delete a row",    "Del");
    window(41,2,screenRight-2,screenBottom-1);
    i = 27;
    gotoxy(1,3);
    cputs("Editing keys\n\n\r");
    PutPair(i,"insert/overwrite mode",    "Ins");
    PutPair(i,"right one character",      "right_arrow");
    PutPair(i," left one character",      "left_arrow");
    PutPair(i,"first character",          "Home");
    PutPair(i," last character",          "End");
    PutPair(i,"delete active character",  "Del");
    PutPair(i,"delete previous character","Backspace\n");
    PutPair(i,"  cuT a row",    "Ctrl + T");
    PutPair(i," Copy a row",    "Ctrl + C");
    PutPair(i,"Paste onto row", "Ctrl + P");
    window(3,2,screenRight-2,screenBottom-1);
    i = 50;
    gotoxy(1,18);
    cputs("Acronym export/import editing keys\n\r");
    PutPair(i,"   cuT-out an acronym (leave empty = null)","T");
    PutPair(i,"   Paste over (replace) an acronymn from cut/del","P");
    PutPair(i,"   Delete = cut then scroll others up",     "D or Del");
    PutPair(i,"   Insert = add null by scroll others down","I or Ins");
}  /* end HelpEdit */

#endif