/* GDFtrans
      data transfer functions
*/

#ifndef GDFtran

int  GDFtran;


/* prototypes internal */


int  GetRT(struct DEFN *Defn, char *textline, char *rectyp);
     /* determine the record type name <rectyp> of record <textline> */
int  TransOpen(char *dataname, char *title, char form, char useFile);
     /* open a data file */
int  cputText(char *text);
     /* write text to screen, allowing for wrapped lines */
char *ReadWriteFile(int lines);
     /* read & write <lines> file records to screen */
int  CheckFormat(struct RECTYPE *RT, const char *textptr, int showError);
     /* checks <text> against the acronym formats
        returns the quantity of discrepancies */
int  outValue(struct ACRONYM *Asource, struct ACRONYM *Adest, char *source);
     /* output <As.width> characters from <source> starting at <As.start>
        making sure <Ad.width> characters are sent out to destStream */
char *ReadVerifyFile(int *lines);
     /* read & verify file records */
int  PickFile(char *title, char *name, char useFile);
     /* select and open a file called <name> */
int  GetFile(char *title, char *name, char useFile);
     /* get data file name and open it for reading */
void TickChoice(int list, int row);
     /* fills the manipulation pick list for item <choice> */
void PrintError(int line, long error, int note);
     /* prints error message for CreateFile */
int  FillAcroFormat(int list, int column, struct RECTYPE *RT);
     /* fills <list> <column> from <RT> with acronym & format */
void FillConvertTable(struct RECTYPE *RTsource, struct RECTYPE *RTdest);
     /* enters the data conversion request into the data structure */


/* prototypes external */


void DisplayData(void);
     /* display a data file */
void VerifyData(void);
     /* verify a data file */
void CreateFile(int gdfImport);
     /* create a data file from one or more source files
        <gdfImport> = 0 for GDF export
                    = 1 for GDF import */


/*  IMPLIMENTATION - internal */

char *textline;                /* to hold a record from the data set */
int   linelen = 4000;          /* max length of a data record */
char  dataname[PATHSIZE] = ""; /* file path+name of current data set */
long  recordNumber;            /* number of data records read */
long  DrecordNumber;           /* number of data records written */
long  DbadNumber;              /* number of bad data records written */
int   useDefn;                 /* = 1 if must display via definition */
int   displayRT;               /* last RT displayed */
int   outCheck = 0;            /* output manipulation flags */


int GetRT(struct DEFN *Defn, char *textline, char *rectyp)
     /* determine the record type name <rectyp> of record <textline> */
{   int R;
    strncpy(rectyp,textline,LENRT-1);       /* extract RT name */
    rectyp[LENRT-1] = '\0';
       /* find the corresponding RT */
    R = 0;
    while ((R < Defn->RTused)
       && (strcmp(RT_ptr[Defn->RTid[R]]->rt,rectyp))) R++;
       /* fix for absent RT option */
    if ((Defn->RTnull != -1) && (R == Defn->RTused)) {
      R = Defn->RTnull;
      strcpy(rectyp,"null");
    }
    return R;
}


int TransOpen(char *dataname, char *title, char form, char useFile)
     /* open a data file
        dataname = returned filename
        title    = filelist header
        form G   = .GDF files
             F   = .GDF extn but ask for a dataname
             E   = extension supplied in dataname
             M   = filename or mask supplied in dataname
             A   = any file
             X   = name supplied in dataname
            none = ask for a filename
        useFile  = file stream [S=source, D=destination, else general] */
{   char extn[PATHSIZE];
    char comment[30];
    int  access;            /* OpenFile access flag */
       /* set up filename extension */
    switch (form) {
      case 'F' : ;
      case 'G' : strcpy(extn,GDFEXTN); break;
      case 'M' : if (strcmp(dataname,ValidateDOSfileName(dataname,'\0')) == 0) {
                   extn[0] = '\0';
                   form = 'X';
                   break;
                 }
      case 'E' : strcpy(extn,dataname); break;
      case 'A' : strcpy(extn,".*"); break;
      default  : extn[0] = '\0';
    }
       /* get user selected file */
    if ((extn[0] != '\0') && (form != 'F')) {  /* from directory list */
      strcpy(dataname,FileList(title, extn,0));
      if (strcmp(dataname,"?") == 0) {
        alert('A',"There are no data files to open !","");
        dataname[0] = '\0';
      }}
    else if (form != 'X') {                    /* from user */
      switch (useFile) {              /* set up filename question */
        case 'S' : strcpy(comment,"need a SOURCE filename"); break;
        case 'D' : strcpy(comment,"need a DESTINATION filename"); break;
        default  : strcpy(comment,"need a filename");
      }
      strcpy(dataname,alert('S',comment,"please enter a name : "));
    }

    if (dataname[0] == '\0')          /* no valid file selected */
      return 0;
    else {
         /* add extn */
      if ((form == 'F') && (strchr(dataname,'.') == NULL)) strcat(dataname,extn) ;
         /* open file */
      access = 2;
      switch (useFile) {         /* set read/write flag */
        case 'D' : strcpy(extn,"w"); access = 1; break;
        case 'S' : ;
        default  : strcpy(extn,"r");
      }
      if (OpenFile(dataname,extn,access,useFile)) {
           /* test for an empty file */
        switch (useFile) {
          case 'D' : if (!feof(destStream)) return 1; break;
          case 'S' : if (!feof(sourceStream)) return 1; break;
          default  : if (!feof(stream)) return 1;
        }
        alert('A',"You have accessed an EMPTY data file called ",dataname);
      }
      return 2;
    }
}  /* end TransOpen */


int cputText(char *text)
     /* write text to screen, allowing for wrapped lines */
{   int I = 1;          /* number of lines used in display */
    if (wrap) {         /* use multiple lines if needed */
        I += strlen(text)/screenRight;
        cputs(text);
      }
    else {              /* no wrap = trim to one line */
      text[screenRight]   = '\n';
      text[screenRight+1] = '\0';
      cputs(text);
    }
    putch('\r');
    return I;
}

char *ReadWriteFile(int lines)
     /* read & write <lines> file records to screen */
{   char *endOF;               /* NULL if eof */
    char rectyp[LENRT];        /* RT in data record */
    char dataInfo[256];        /* data belonging to an acronym */
    char *textptr;             /* pointer into textline */
    int  I,R,A,L;              /* counters */
    int  X,Y;                  /* screen column,row */
    static int scrnWidth;      /* logical screen width */
    int  Nlines = 1;           /* number of lines per record */

    SetTextColour(BLACK,WHITE);
    for (I = 0; I < lines; I++) {
      if ((endOF = fgets(textline,linelen-1,stream)) == NULL)  /* read file */
        I = lines;                                             /* skip if EOF */
      else {
        recordNumber++;
        textline[strlen(textline)+1] = '\0';       /* add extra null */
        if (useDefn) {                             /* formatted dump */
          R = GetRT(DefnActive, textline, rectyp); /* get RT index & name */
             /* print RT if changed */
          if (R != displayRT) {
            I++;                                  /* allow room for RT line */
            if (R < DefnActive->RTused) {
              RT = RT_ptr[DefnActive->RTid[R]];
              if (wrap) {
                scrnWidth = 1 + RT->order;
                Nlines = 1 + RT->order/screenRight;
              }
              else {
                scrnWidth = screenRight;          /* trim to screen width */
                Nlines = 1;
              }
              sprintf(dataInfo,"\nRecord type of %s",rectyp);

              PrintCharWindow('l',gety(),WHITE,BLACK, dataInfo); /* print RT */
              putch('\n');
              I += Nlines;                        /* allow room for Acro lines */
                 /* print Acronym list */
              for (A = 0, L = Acronym_ptr[RT->Aid[0]]->fwidth;
                   (A < RT->Aused) && (L <= scrnWidth); A++) {
                Acro = Acronym_ptr[RT->Aid[A]];
                sprintf(dataInfo,"%-*s",Acro->order-1,Acro->acronym);  /* padding */
                SetTextColour(RED,WHITE);
                cputs(dataInfo);
                SetTextColour(WHITE,BLACK);
                putch(' ');
                L += Acro->order;
              }
              SetTextColour(BLACK,WHITE);
            }
            else {
              PrintCharWindow(253,gety(),WHITE,BLACK, "\nnot a recognised RT !");
              Nlines = 1;
            }
            putch('\n');
            displayRT = R;
          }
             /* print data */
          if (displayRT < DefnActive->RTused) {
            textptr = textline;
            if (DefnActive->RTnull != displayRT) textptr += 4;  /* skip RT */
            Y = gety();
            for (A = 0, X = 1;
                (textptr[0] != '\0') && (A < RT->Aused) && (X <= scrnWidth);
                 A++) {
              Acro = Acronym_ptr[RT->Aid[A]];
              strncpy(dataInfo,textptr,Acro->fwidth);
              dataInfo[Acro->fwidth] = '\0';
              dataInfo[screenRight-X+1] = '\0';
              gotoxy(X % screenRight,Y);
              cputs(dataInfo);                    /* write piece of record */
              X += Acro->order;                   /* add display shift */
              if (strlen(textptr) > Acro->fwidth) /* shift text pointer */
                textptr += Acro->fwidth;
              else
                textptr[0] = '\0';
            }
               /* display leftovers */
            if (textptr[0] != '\0') I += cputText(textptr) -1;
          }
          else            /* error dump */
            I += cputText(textline) -1;
        }
        else {            /* text dump */
          I += cputText(textline) -1;
        }
      }
      ClearLine(BLACK,BLACK,WHITE);     /* removes "white-out" on scroll */
    }

    if (endOF == NULL) {
      PrintCharScreen('L',screenBottom, WHITE,BLACK,"End-of-File at record");
      X = 23;
      }
    else
      X = 8;

    PrintIntScreen(X,screenBottom, WHITE,BLACK, recordNumber);

    return endOF;
}  /* end ReadWriteFile */


int CheckFormat(struct RECTYPE *RT, const char *textptr, int showError)
     /* checks <text> against the acronym formats
        returns the quantity of discrepancies */
{   int A,I;         /* indexes */
    char info[256];  /* test space */
    int error = 0;   /* error flag */
       /* format test dummies */
    int      testI;
    double   testD;
    long int testL;
    char     testS[200];

    for (A = 0; A < RT->Aused; A++) {
      Acro = Acronym_ptr[RT->Aid[A]];
      strncpy(info,textptr,Acro->fwidth);   /* extract data segment */
      info[Acro->fwidth] = '\0';
      switch (Acro->format[0]) {            /* test format against data */
        case 'I' : I = sscanf(info,"%d%s",&testI,&testS); break;
        case 'L' : I = sscanf(info,"%D%s",&testL,&testS); break;
        case 'F' : I = sscanf(info,"%lf%s",&testD,&testS); break;
        case 'E' : ;
        case 'D' : ;
        case 'G' : I = sscanf(info,"%le%s",&testD,&testS); break;
        case 'X' : ;
        case 'A' : I = 1; break;
        default  : I = 0;
      }
      if (I != 1) {                         /* error in format */
        if (showError) {
          PrintCharWindow(253,gety(),RED,WHITE, "format error in record ");
          PrintInt(getx(),gety(),"%d",recordNumber);
          PrintChar(getx(),gety()," at item %s\r\n",Acro->acronym);
        }
        error += 1;
      }
      textptr += Acro->fwidth;
    }
    return error;
}  /* end CheckFormat */


int outValue(struct ACRONYM *Asource, struct ACRONYM *Adest, char *source)
     /* output <As.width> characters from <source> starting at <As.start>
        making sure <Ad.width> characters are sent out to destStream */
{   int P;             /* indexes */
    int done = 0;      /* not output */
    int Ntype;         /* numeric type - integer or float */
    char form;         /* format type on read of source */
    char fd;           /* format type of destination */
    int widths,widthd; /* format width of source and destination */
    double   testD;    /* format dummies */
    long int testL;

    char info[21];     /* an extract from source */
    char *sptr;        /* pointer into source */

    sptr = source + Asource->fstart;
    form = Asource->format[0];
    fd   = Adest->format[0];
    widths = Asource->fwidth;                         /* fix up for short record */
      P = strlen(sptr); if (P < widths) widths = P-1; /* allow for '\n' */
    widthd = Adest->fwidth;

       /* blank input or output */
    if ((form == 'X') || (fd == 'X')) {
      for (P = 0; P < widthd; P++) fputc(' ',destStream);
      done = 1;
    }
       /* identical format types OR A output */
    else if ((form == fd) || (fd == 'A')) {
      if (widths > widthd) {        /* trim leading blanks */
        for (P = 0; P < widths-widthd, sptr[P] == ' '; P++);
        sptr += P;
        widths -= P;
        if ((widths > widthd) && (form != 'A')) {  /* trim trailing blanks */
          for (P = widths; P > widths-widthd, sptr[P] == ' '; P--);
          widths -= P;
          if (widths <= widthd) done = 1;
          /* else process numerically */
        }
        else done = 1;
      }
      else done = 1;

      if (done == 1) {
        for (P = 0; P < widths; fputc(sptr[P++],destStream));  /* transfer */
        for (P = P; P < widthd; P++) fputc(' ',destStream);    /* pad */
      }
    }
       /* text source, numeric destination */
    else if (form == 'A') form = fd;
       /* else numeric source */


    if (done == 0) {
      strncpy(info,sptr,widths);   /* extract data segment */
      info[widths] = '\0';
      switch (form) {              /* read value */
        case 'I' : ;
        case 'L' : done = sscanf(info,"%D",&testL); Ntype = 0; break;
        case 'F' : ;
        case 'E' : ;
        case 'D' : ;
        case 'G' : done = sscanf(info,"%le",&testD); Ntype = 1;
      }
      if (done != 1)    /* error in read */
        for (P = 0; P < widthd; P++) fputc(' ',destStream);
      else {
        switch (fd) {           /* write value */
          case 'I' : ;
          case 'L' : if (Ntype == 1) testL = testD;
                     fprintf(destStream,Adest->Cformat,testL);
                     break;
          case 'F' : ;
          case 'E' : ;
          case 'D' : ;
          case 'G' : if (Ntype == 0) testD = testL;
                     fprintf(destStream,Adest->Cformat,testD);
        }
        done = 1;
      }
    }
    if (done == 1) return 0; else return 2;
}  /* end outValue */


char *ReadVerifyFile(int *lines)
     /* read & verify file records */
{   char *endOF;               /* NULL if eof */
    char *textptr;             /* pointer into textline */
    char rectyp[LENRT];        /* RT in data record */
    int  I,R,A,L;              /* counters */
    static int scrnWidth;      /* effective screen width */
    int  Nlines = 1;           /* number of lines per record */

       /* read and verify each record */
    do if ((endOF = fgets(textline,linelen-1,stream)) != NULL) {  /* read file */
      recordNumber++;
      textline[strlen(textline)+1] = '\0';      /* add extra null */
      R = GetRT(DefnActive, textline, rectyp);  /* get RT index & name */

      if (R != displayRT) {                     /* RT changed */
        ClearLine(BLACK,WHITE,BLACK);           /* set colours */
        *lines += 2;                            /* RT name lines */
        if (R >= DefnActive->RTused) {
          PrintInt(1,gety(),"\nNot a recognised RT starting at record %d\r\n",recordNumber);
          Nlines = 1;
          }
        else {
          RT = RT_ptr[DefnActive->RTid[R]];
          PrintChar(1,gety(),"\nRecord type of %s",rectyp);
          PrintInt(getx(),gety()," with length %d",RT->order);
          PrintInt(getx(),gety()," starting at record %d\r\n",recordNumber);
          if (wrap) {
            scrnWidth = 1 + RT->order;
            Nlines = 1 + RT->order/screenRight;
            *lines += Nlines-1;
          }
          else {
            scrnWidth = screenRight;
            Nlines = 1;
          }
             /* print Acronym list */
          for (A = 0, L = Acronym_ptr[RT->Aid[0]]->fwidth;
              (A < RT->Aused) && (L <= scrnWidth); A++) {
            Acro = Acronym_ptr[RT->Aid[A]];
            SetTextColour(LIGHTGRAY,BLUE);
            if (Acro->fwidth < Acro->order-1) {  /* highlight part */
              for (I = 0; I < Acro->fwidth; I++) putch(Acro->acronym[I]);
              SetTextColour(BLUE,LIGHTGRAY);
              for (I = Acro->fwidth; I < Acro->order-1; I++) putch(Acro->acronym[I]);
              SetTextColour(LIGHTGRAY,BLUE);
              }
            else {                              /* highlight all */
              sprintf(buff,"%-*s",Acro->order-1,Acro->acronym);
              cputs(buff);
            }
            SetTextColour(WHITE,BLACK);
            putch(' ');
            L += Acro->order;
          }
             /* fix up display */
          L = WhereY();
          I = 2;                                /* menu + header */
          SetWindow(0,I,0,2);                   /* not top and bottom lines */
          gotoxy(1,L-I);                        /* keep row static */
          putch('\n');putch('\n');
          *lines += I;                          /* header lines */
        }
        displayRT = R;
      }
         /* test data */
      SetTextColour(BLACK,WHITE);
      if (displayRT >= DefnActive->RTused)      /* not a valid RT */
        *lines += cputText(textline);
      else if (strlen(textline) != RT->order) { /* wrong record length */
        PrintCharWindow(253,gety(),RED,WHITE, "inconsistant with defined record length at record ");
        PrintInt(getx(),gety(),"%d\r\n",recordNumber);
        ClearLine(BLACK,BLACK,WHITE);
        *lines += 1 + cputText(textline);
        }
      else {                                    /* test each acronym */
        textptr = textline;
        if (DefnActive->RTnull != displayRT) textptr += 4;  /* skip RT */
        I = CheckFormat(RT, textptr, 1);
        if (I) {
          ClearLine(BLACK,BLACK,WHITE);
          *lines += I + cputText(textline);
        }
      }
      ClearLine(BLACK,BLACK,WHITE);
      PrintIntScreen(8,screenBottom, WHITE,BLACK, recordNumber);

    } while ((endOF != NULL) && (*lines < screenBottom-4-Nlines));
    if (endOF == NULL) {
      PrintCharScreen('L',screenBottom, WHITE,BLACK, "End-of-File at record");
      PrintIntScreen(23,screenBottom, WHITE,BLACK, recordNumber);
    }
    return endOF;
}  /* end ReadVerifyFile */


int PickFile(char *title, char *name, char useFile)
     /* select and open a file called <name> */
{   char form;           /* type of file */
    int pick;            /* index to selected form */
    int list =    5;     /* display list index */
    int left =   47;     /* position of list on screen */
    int top  =    2;
    char temp[10];       /* work area */
       /* ask for file type */
    ListNew(list,"FILENAME or MASK of");
    ListFillInit(list,0,title,TEXTTYPE,1,0,13,0);
    temp[0] = '*'; temp[1] = '\0'; strcat(temp,GDFEXTN); /* = *.GDF */
    ListFill(list,0,0,"*.*",0);           /* any file */
    ListFill(list,0,1,temp,0);            /* GDF-type file */
    ListFill(list,0,2,dataname,0);        /* current data file */
    ListFill(list,0,3,dataSource,0);      /* last data input file */
    ListFill(list,0,4,dataDestination,0); /* last data output file */
    ListFill(list,0,5,"",0);              /* allow user-specified */
    Lcolt = WHITE;
    Lcolb = BLACK;
    pick = ListEdit(list, left, top, 0, 0, 0);
    if (pick == -1) return 0;             /* Esc pressed */
    else {                                /* filename selected or typed */
      form = 'M';
      Lcolm = listSet[list]->List[0];
      ListGetString(Lcolm, pick, name);
      if (name[0] == '\0') form = ' ';    /* no filename - ask for one ! */
    }
    ListColumnsDispose(list);             /* dispose of list */
    return TransOpen(name,"AVAILABLE FILES",form,useFile);
}  /* end PickFile */

int GetFile(char *title, char *name, char useFile)
     /* get data file name and open it for reading */
{
    if (DefnActive == NULL) {             /* no definition active */
      alert('C',"You must write or select a Definition first","press a key to ABORT");
      return 0;
      }
    else  /* get data file name and open it for reading */
      return PickFile(title,name,useFile);      /* file selected */
}  /* end GetFile */

void TickChoice(int list, int row)
     /* fills the manipulation pick list for item <choice> */
{   char text[30];           /* option message */
    char tick[3] = " ";     /* flag for picked or not */
    text[0] = tick[(  ((row/2 == 0) &&  (outCheck %  4 == row+1))
                   || ((row   == 2) && ((outCheck & 12)  >  0))
                   || ((row   == 3) && ((outCheck &  8)  >  0))
                   || ((row   == 4) && ((outCheck & 16) == 16)))];
    text[1] = '\0';
    switch (row) {
      case 0 : strcat(text," ADD active record type"); break;
      case 1 : strcat(text," REMOVE record type"); break;
      case 2 : strcat(text," VERIFY record length"); break;
      case 3 : strcat(text," VERIFY formats"); break;
      case 4 : strcat(text," CONVERT formats"); break;
    }
    ListFill(list,0,row,text,0);
}  /* end TickChoice */

void PrintError(int line, long error, int note)
     /* prints error message for CreateFile */
{   static char message[4][14]
      = { "wrong length"
         ,"bad format"
         ,"conversion"
         ,"RT unknown" };

    PrintInt(5,line+1,"%d ERRORS;",error);
    PrintInt(getx(),gety()," record %d :",recordNumber);
    PrintChar(getx(),gety()," %-12s\a",message[note-1]);
    DrecordNumber--;
}  /* end PrintError */


int FillAcroFormat(int list, int column, struct RECTYPE *RT)
     /* fills <list> <column> from <RT> with acronym & format */
{   int  A;
    char tempstr[21];        /* a string dummy */
       /* fill with acronym+formats */
    for (A = 0; A < RT->Aused; A++) {
      Acro = Acronym_ptr[RT->Aid[A]];
      sprintf(tempstr,"%-10s %s", Acro->acronym,Acro->format);
      ListFill(list,column,A,tempstr,0);
    }

    return A;
} /* end FillAcroFormat */


void FillConvertTable(struct RECTYPE *RTsource, struct RECTYPE *RTdest)
     /* enters the data conversion request into the data structure */
{   int I,A;

       /* source to destination RT conversion */
    RTsource->reference = RTdest;
       /* fill acronym cross-references */
    for (A = 0; A < RTdest->Aused; A++) {
      Acro = Acronym_ptr[RTdest->Aid[A]];
      I = ListGetOriginal(A);               /* original row */
      if (I >= 0) {
        Acro->reference = Acronym_ptr[RTsource->Aid[I]];
        /* *** poor conversion warning *** */
      }
      else Acro->reference = NULL;
    }
}  /* end FillConvertTable */


/*  IMPLIMENTATION - external */


void DisplayData(void)
     /* display a data file */
{
    int  userkey;           /* key pressed by user */
    char *endDisplay;       /* NULL if eof or quit display */
    int  quit = 0;          /* true if to quit */

       /* get data file name and open it for reading */
    if (PickFile("Display file",dataname,'S') == 1) {
         /* file exists */
      if (DefnActive == NULL)       /* no definition active */
        useDefn = 0;
      else
        useDefn = alertYesNo("Do you want to use the active Definition ?","Yes/No ");

      textline = (char *) malloc(linelen);
      recordNumber = 0;
      if (!useDefn)                 /* no formating */
        ClearTick(10,WHITE,BLACK);  /* clear screen and make tick marks */
      else {                        /* use active definition */
        ClearTick(0,WHITE,BLACK);   /* clear screen */
        ClearWindow();              /* clear screen under menu */
        FormatWidths();             /* establish acronym column sizes */
        displayRT = -1;             /* = no current RT on display */
      }
      PrintCharScreen('C',screenBottom, BLACK, WHITE, dataname);
      PrintCharScreen('L',screenBottom, WHITE,BLACK, "record");

         /* read & write file records to screen */
      endDisplay = ReadWriteFile(screenBottom-5);
         /* user request loop */
      do {
        userkey = getch();                          /* wait for user response */
        if (endDisplay == NULL) {                   /* quit */
          if (userkey == EXTKEY) userkey = getch(); /* discard double keys */
          quit = 1;
          }
        else switch (userkey) {                     /* do what user requested */
          case EXTKEY : userkey = getch();          /* get double key */
            switch (userkey) {
              case DNARROW : endDisplay = ReadWriteFile(1);
                             break;
              case PAGEDN  : SetTextColour(WHITE,BLACK);
                             ClearWindow();
                             SetTextColour(BLACK,WHITE);
                             endDisplay = ReadWriteFile(screenBottom-5);
                             break;
              default      : putchar(BELL);  /* a do-nothing key */
            }
            break;
          case ESC    : quit = 1; break;
          case ENTER  : endDisplay = ReadWriteFile(1);
                        break;
          default     : putchar(BELL);       /* a do-nothing key */
        }
      } while (!quit);
      free(textline);
    }
}  /* end DisplayData */


void VerifyData(void)
     /* verify a data file */
{
    int  userkey;           /* key pressed by user */
    char *endDisplay;       /* NULL if eof or quit display */
    int  lines = 0;         /* [error] lines displayed */

    if (GetFile("Verification file",dataname,' ') == 1) {  /* file and definition exist */
      textline = (char *) malloc(linelen);
      recordNumber = 0;
      FormatWidths();                 /* establish acronym column sizes */
      displayRT = -1;                 /* = no current RT on display */
      ClearTick(0,WHITE,BLACK);       /* clear screen */
      PrintCharScreen('C',screenBottom, BLACK, WHITE, dataname);
      PrintCharScreen('L',screenBottom, WHITE,BLACK, "record");

         /* read & verify file records */
      do {
        endDisplay = ReadVerifyFile(&lines);
        userkey = getch();                      /* wait for user response */
        if (endDisplay == NULL) {               /* quit */
          if (userkey == EXTKEY) userkey = getch(); } /* discard double keys */
        else switch (userkey) {                 /* do what user requested */
          case EXTKEY : userkey = getch();      /* get double key */
            switch (userkey) {
              case DNARROW : lines -= 10; break; /* allow up to 5 errors */
              case END     : ;
              case PAGEDN  : lines = 0; break;   /* allow new set of errors */
              default      : putchar(BELL);         /* a do-nothing key */
            }
            break;
          case ESC    : endDisplay = NULL;   /* quit */
          case ENTER  : lines -= 3; break;   /* allow another error */
          default     : putchar(BELL);          /* a do-nothing key */
        }
      } while (endDisplay != NULL);
      free(textline);
    }
}  /* end VerifyData */


void CreateFile(int gdfImport)
     /* create a data file from one or more source files
        <gdfImport> = 0 for GDF export
                    = 1 for GDF import */
{
    int  list =  4;          /* adjustments list number */
    int  left = 47;          /* position of list on screen */
    int  top  =  2;
    char rtnameS[5];         /* source record-type name */
    char rtname[5];          /* record-type name, 4-char long */
    int  rtlength;           /* length of the active RT */
    long errors;             /* count of record check failures */
    int  userkey;            /* key pressed by user */
    int  userinfo;           /* window row for passing info to the user */
    char *textptr;           /* pointer into textline */
    int  RToffset;           /* allowance for RT in data source */
    int  A,R,I,K;            /* loop index, calc dummies */
    int  row;                /* a List row */
    int  messNumber;         /* error message number */
    char defnSource[PATHSIZE]; /* source file definition filename */
    struct RECTYPE *RTsource, *RTdest;   /* RT dummies */
    struct DEFN *DefnSource,*DefnDestination; /* source & destination defns */

       /* ask for DESTINATION file */
    if (TransOpen(dataDestination,"",'F','D') == 1) { /* get filename & open file */
      textline = (char *) malloc(linelen);
      ClearTick(0,WHITE,BLACK);   /* clear screen */
      PrintChar(1,2,"DESTINATION file %s", dataDestination);
      DrecordNumber = 0;
      DbadNumber = 0;
      PrintInt(3,3,"records written %d",DrecordNumber);
      userinfo = 3;
         /* make a DFN file of active definition */
      DefnDestination = DefnActive;
      if (!gdfImport) MakeDefn(-1);
         /* establish Destination acronym column sizes */
      FormatWidths();
         /* ask for first source file */
      if (GetFile("first SOURCE file",dataSource,'S') == 1) {  /* file and definition exist */
           /* set user-selected RT as default */
        RT = RT_ptr[active[RTLIST]];
           /* set up check/conversion list */
        Lcolt = WHITE;
        Lcolb = BLACK;
        ListNew(list,"SOURCE to DESTINATION");
        ListFillInit(list,0," adjustments selected",TEXTTYPE,1,0,28,0);
        for (I = 0; I < 5; I++) TickChoice(list,I);

           /* start data read/write loop */
        do {
          errors = 0;
           /* ask for check/conversion option */
          A = -1;
          do {
            A += 1; if (A > 4) A = 0;                 /* active row of list */
            A = ListEdit(list, left,top, 0, A, 2);    /* 2 = use list as a menu */
            switch (A) {                              /* toggle */
              case 0 : ;                           /* add/remove RT */
              case 1 : outCheck = (((29+A) & outCheck) ^ (A+1));
                       for (I = 0; I < 2; I++) TickChoice(list,I);
                       break;
              case 2 : ;                           /* length/format */
              case 3 : outCheck = (((15+4*A) & outCheck) ^ (4*A-4));
                       for (I = 2; I < 4; I++) TickChoice(list,I);
                       break;
              case 4 : outCheck ^= 16;             /* conversion */
                       TickChoice(list,4);
                       break;
            }
          } while (A != -1);
             /* set up default verification parameters */
          if (outCheck % 2 == 1) {
            strcpy(rtname,"    ");       /* fill with 4-blanks */
            for (I = 0; I < strlen(RT->rt); I++) rtname[I] = RT->rt[I];
            rtlength = RT->order - 4;    /* less 4 for RT */
            RToffset = 0;
            }
          else {
            rtlength = RT->order;
            RToffset = 4;
          }
             /* format conversion table */
          if (outCheck < 16)                 /* no change in format */
            DefnSource = DefnDestination;
          else {                             /* re-order items or format conversion */
               /* load source definition */
            strcpy(defnSource,ValidateDOSfileName(dataSource,'.'));
            GetDefn(defnSource);
            if (defnSource[0] != '\0')       /* DFN file to match source */
              DefnSource = DefnActive;
            else {                           /* no matching .DFN file */
                 /* pick from LIB */
              ListNew(++list,"LIBRARY DEFINITION of");
              ListFillInit(list,0,dataSource,TEXTTYPE,1,0,9,0);
              for (DefnSource = DefnRoot, row = 0; DefnSource != NULL;
                   DefnSource = DefnSource->next, row++)
                ListFill(list,0,row,DefnSource->defn,0);
              ListFill(list,0,row,"use a DFN file",0);
              A = ListEdit(list,left,top, 0, row, 2); /* use list as a menu */
              ListColumnsDispose(list--);             /* dispose of list */
              if (A == row) {                         /* pick from .DFN */
                strcpy(defnSource,FileList("DEFINITIONS",DFNEXTN,0));
                if (defnSource[0] == '\0') A = -1;    /* rejected all */
                else {
                  GetDefn(defnSource);
                  DefnSource = DefnActive;
                }
              }
              else if (A >= 0)                        /* picked a LIB defn */
                for (DefnSource = DefnRoot, I = 0; I != A;
                     DefnSource = DefnSource->next, I++);
            }
            FormatWidths();                           /* establish source format widths */
            DefnActive = DefnDestination;
            if (A == -1) outCheck ^= 16;              /* toggle off */
               /* generate a conversion table for each source record type */
            else {
                 /* make RT menu */
              ListNew(++list,"SOURCE RT -->");
              ListFillInit(list,0,"Destination RT",TEXTTYPE,1,0,5,0);
              for (I = 0; I < DefnDestination->RTused; I++)
                  ListFill(list,0,I,RT_ptr[DefnDestination->RTid[I]]->rt,0);

              for (R = 0; R < DefnSource->RTused; R++) {
                RTsource = RT_ptr[DefnSource->RTid[R]];
                   /* make acronymn list */
                ListNew(++list,"ACRONYM CONVERSION");
                ListFillInit(list,0,"Destination",TEXTTYPE,1,0,15,0);
                sprintf(textline,"Source = %s",RTsource->rt);
                ListFillInit(list,1,textline,TEXTTYPE,1,0,15,0);
                FillAcroFormat(list, 1, RTsource);
                SetColumnMin(list, 1);                     /* only edit source items */
                ListOn(list,0, 24,3, 'S', Lcolt, Lcolb);   /* display list */
                   /* select destination RT */
                A = 0;
                do {
                     /* (re-)fill destination list */
                  RTdest = RT_ptr[DefnDestination->RTid[A]]; /* user choice */
                  FillAcroFormat(list, 0, RTdest);
                  for (I = RTdest->Aused; I < Lrecd->qtyInList; ListFill(list,0,I++," ",0));
                  ListRedisplay(list, 0,0);
                     /* user RT selection */
                  A = ListEdit(list-1, 5,3, 0, A, 3);
                } while (A >= 0);

                   /* RT selected - clear single-lined acro list */
                ListOff(list);
                   /* blank fill any spare rows */
                for (I = RTsource->Aused; I < Lrecd->qtyInList; ListFill(list,1,I++," ",0));
                   /* now edit conversion */
                A = ListEdit(list, 24,3, 1, RTsource->Aused, 4);
                   /* extract convertors then dispose of list */
                FillConvertTable(RTsource,RTdest);
                ListColumnsDispose(list--);  /* acronym conversion list */
              }
              ListColumnsDispose(list--);  /* RT conversion list */
            }
          }
             /* reset window */
          if (userinfo > screenBottom-5) userinfo = 5;
          else userinfo += 2;
          SetWindow(0,2,0,1);  /* full screen less top & bottom lines */
          SetTextColour(WHITE,BLACK);
             /* show user what they asked for */
          PrintChar(1,userinfo++,"SOURCE file %s",dataSource);
          recordNumber = 0;
          PrintInt(5,gety()+1,"records read %d",recordNumber);
          K = outCheck/4;              /* verification type */
          messNumber = 0;              /* no message */
          displayRT = -1;              /* no previous rt read */
          userkey = 0;
            /* read Source record */
          while ((userkey != ESC) && (fgets(textline,linelen-1,sourceStream) != NULL)) {
            recordNumber++;
               /* write Destination record */
            DrecordNumber++;
            textptr = textline;
            if (K < 4) {                                /* no format change */
              if (outCheck % 4 == 1)                    /* add default RT */
                fputs(rtname,destStream);
              else if (outCheck % 4 == 2) textptr += 4; /* remove RT */
            }
            if (K > 0) { /* need checks */
                 /* get source RT */
              R = GetRT(DefnSource, textline, rtnameS);
              if (R != displayRT) {                     /* rt of data changed */
                displayRT = R;
                if (R >= DefnSource->RTused) {
                  messNumber = 4;
                  rtlength = 0;
                  }
                else {
                  RTsource = RT_ptr[DefnSource->RTid[R]];
                  RT = RTsource;
                  if (K > 3) {
                    RTdest = RTsource->reference;
                    if (RTdest == NULL)         /* if NULL then reject */
                      rtlength = 0;
                    else
                      rtlength = RTsource->order;
                  }
                  else
                    rtlength = RTsource->order;
                }
                if (outCheck % 2 == 1) rtlength -= 4; /* less 4 for missing RT */
              }
                         /* check record length */
              if (rtlength > 0) {   /* accepted source record */
                if (strlen(textline) != rtlength)
                  messNumber = 1;
                else if (K > 1) {
                        /* check format */
                  if (((K == 2) || (K == 6)) && (CheckFormat(RT, textline+RToffset, 0)))
                    messNumber = 2;
                  else if (K > 3) {                   /* convert format */
                      /* output RT name */
                    if (outCheck % 4 != 2) {
                      if (strcmp(RTdest->rt,"<>"))    /* try dest RT */
                        fputs(RTdest->rt,destStream);
                      else if (strcmp(rtnameS,"<>"))  /* try source RT */
                        fputs(rtnameS,destStream);
                      else                            /* use default RT */
                        fputs(rtname,destStream);
                    }
                       /* output data */
                    for (A = 0; A < RTdest->Aused; A++) {
                      Acro = Acronym_ptr[RTdest->Aid[A]];
                      AcroS = Acro->reference;
                      if (AcroS == NULL) {            /* no source acronym */
                        if (Acro->null == NOVALUE)    /* no default to use */
                          fprintf(destStream,"%*s",Acro->fwidth," ");
                        else {                        /* use default */
                          if ((Acro->format[0] == 'I') || (Acro->format[0] == 'L'))
                            fprintf(destStream,Acro->Cformat,(long int)Acro->null);
                          else
                            fprintf(destStream,Acro->Cformat,Acro->null);
                        }}
                      else {
                        I = outValue(AcroS,Acro, textptr);
                        if (messNumber == 0) messNumber = I;
                      }
                    }
                    fputc('\n',destStream);
                    if (messNumber > 0) DbadNumber++;
                  }
                }
              }
            }
               /* display error or write data */
            if (messNumber > 0) {
              PrintError(userinfo,++errors,messNumber);
              messNumber = 0;
              }
            else if (K < 4)
              fputs(textptr,destStream);

               /* allow user interrupt */
            if (kbhit) {
              userkey = getch();            /* get user key */
              if (userkey == EXTKEY) userkey = getch(); /* discard double keys */
              else if (userkey == ESC)                /* stop transfer */
                alert('A',"User ABORT","");
            }
               /* show progress */
            if (recordNumber % 100 == 0) {
              PrintInt(20,userinfo,"%d",recordNumber);
              PrintInt(19,3,"%d",DrecordNumber);
            }
          }
             /* end of transfer */
          PrintInt(5,userinfo,"End-of-File at record %d",recordNumber);
          PrintInt(19,3,"%d",DrecordNumber);
          if ((K > 3) && (DbadNumber > 0)) PrintInt(3,4,"plus %d BAD records",DbadNumber);
          fclose(sourceStream); sourceStream = NULL;
             /* dispose of defn placed in library */
          if (outCheck > 15) DisposeDefn(DefnSource);
             /* until no more sources selected */
        } while (PickFile("next SOURCE file",dataSource,'S') == 1);
        ListColumnsDispose(list);                     /* dispose of list */
      }
      free(textline);
      DefnActive = DefnDestination;
      fclose(destStream); destStream = NULL;
    }
}  /* end CreateFile */


#endif
